//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - model step/impulse response viewer
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#include <stdlib.h>
#pragma hdrstop

#include "MS_form.h"
#include "sys_model.h"
#include "regulator.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMSForm *MSForm;
//---------------------------------------------------------------------------
__fastcall TMSForm::TMSForm(TComponent* Owner)
	: TForm(Owner)
{
	DoubleBuffered=true;

	// no waveform yet
	y=NULL;
	ylen=0;
}
//---------------------------------------------------------------------------
void __fastcall TMSForm::FormClose(TObject *Sender, TCloseAction &Action)
{
	if(y)
		free((void*)y);
	ylen=0;
}
//---------------------------------------------------------------------------
void __fastcall TMSForm::FormResize(TObject *Sender)
{
	DrawStep();
}
//---------------------------------------------------------------------------
void __fastcall TMSForm::ElenChange(TObject *Sender)
{
	DrawStep();
}
//---------------------------------------------------------------------------
void __fastcall TMSForm::CBstepClick(TObject *Sender)
{
	DrawStep();	
}
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// redraw step/impulse
//---------------------------------------------------------------------------
void TMSForm::DrawStep(void)
{
	int i;

	// get samples count
	double Ts;
	reg_get_T(&Ts);
	double tmax=(double)UDlen->Position;
	int n=tmax/Ts;

	// allocate memory if not long enough
	if(ylen<n || !y)
	{
		y=(double*)realloc((void*)y,n*sizeof(double));
		if(!y)
		{
			y=NULL;
			ylen=0;
			return;
		}
		ylen=n;
	}

	// get model
	double th[SYS_N];
	sys_idn_get_model(NULL,th,NULL,NULL,NULL);

	// simulate model
	double x=1;
	sys_idn_lsim(th,&x,y,n,(CBstep->Checked)?LSIM_STEP:LSIM_PULSE);

	// source image
	Graphics::TBitmap *bmp=img->Picture->Bitmap;
	TCanvas *can=bmp->Canvas;

	// set format
	bmp->Width=img->Width;
	bmp->Height=img->Height;
	bmp->PixelFormat=pf32bit;

	// size
	int xs=bmp->Width;
	int ys=bmp->Height;

	// clear image
	can->Pen->Color=clBlack;
	can->Pen->Style=psSolid;
	can->Brush->Color=clWhite;
	can->Brush->Style=bsSolid;
	can->Rectangle(0,0,xs,ys);
	can->MoveTo(SGR_YAX_X,0);
	can->LineTo(SGR_YAX_X,ys-1);

	// find scale
	double gain=(CBstep->Checked)?sys_idn_get_model_gain():0.0;
	double min=(gain<0)?gain:0.0;
	double max=(gain>0)?gain:0.0;
	for(i=0;i<n;i++)
	{
		if(y[i]<min)
			min=y[i];
		if(y[i]>max)
			max=y[i];
	}
	if(min==0.0 && max==0.0)
	{
		min=-1;
		max=+1;
	}
	double dmm=(max-min)*0.1;
	if(max!=0.0)
		max+=dmm;
	if(min!=0.0)
		min-=dmm;


	// scale and zero offset
	double ss=(double)ys/(max-min);
	int zy=((double)ys*(max)/(max-min));
	if(min==0.0)
	{
		zy=ys-SGR_XAX_Y-1;
		ss=(double)zy/max;
	}
	else if(max==0.0)
	{
		zy=SGR_XAX_Y;
		ss=-(double)(ys-SGR_XAX_Y-1)/min;
	}

	// draw zero axis
	can->Pen->Style=psSolid;
	can->Brush->Style=bsClear;
	can->MoveTo(SGR_YAX_X-SGR_YAX_MX,zy);
	can->LineTo(xs-1,zy);


	// calculate y axis mark step
	int yamn=SGR_YAX_MKN;
	i=0;
	double yaxs=(max-min);
	while(1)
	{
		double tmp;
		if(i%3==0)
			tmp=1.0;
		else if(i%3==1)
			tmp=2.0;
		else
			tmp=5.0;
		tmp*=pow(10,(i/3-3));

		if(yaxs/tmp<=yamn || tmp>1E+9)
		{
			yaxs=tmp;
			break;
		}

		i++;
	}

	// now create y axis marks buffer
	double yaxmark[SGR_YAX_MKN*2];
	yamn=0;
	double tmp=0.0;
	for(i=0;i<SGR_YAX_MKN && max>0.0;i++)
	{
		yaxmark[yamn++]=tmp;
		tmp+=yaxs;
		if(tmp>=max)
			break;
	}
	tmp=0.0;
	for(i=0;i<SGR_YAX_MKN && min<0.0;i++)
	{
		yaxmark[yamn++]=tmp;
		tmp-=yaxs;
		if(tmp<=min)
			break;
	}



	// calculate y axis mark step
	int xamn=SGR_XAX_MKN;
	i=0;
	double xaxs=(double)UDlen->Position;
	while(1)
	{
		double tmp;
		if(i%3==0)
			tmp=1.0;
		else if(i%3==1)
			tmp=2.0;
		else
			tmp=5.0;
		tmp*=pow(10,(i/3-1));

		if(xaxs/tmp<=xamn || tmp>1E+6)
		{
			xaxs=tmp;
			break;
		}

		i++;
	}

	// now create y axis marks buffer
	double xaxmark[SGR_XAX_MKN];
	xamn=0;
	tmp=0.0;
	for(i=0;i<SGR_XAX_MKN && max>0.0;i++)
	{
		tmp+=xaxs;
    if(tmp>=(double)UDlen->Position)
			break;
		xaxmark[xamn++]=tmp;
	}


  // draw marks
	int awd=can->TextWidth("t [s]");
	int ahe=can->TextHeight("t [s]");
	can->Brush->Style=bsClear;
	can->Font=FD->Font;
	for(i=0;i<xamn;i++)
	{
		int xx=SGR_YAX_X+(int)((xs-SGR_YAX_X)/tmax*xaxmark[i]);
		if(xx<SGR_YAX_X) xx=SGR_YAX_X; else if(xx>=xs-awd-1) continue;
		int ya=zy-SGR_XAX_MX;
		if(ya<0) ya=0; else if(ya>=ys) ya=ys-1;
		int yb=zy+SGR_XAX_MX;
		if(yb<0) yb=0; else if(yb>=ys) yb=ys-1;

		for(ya;ya<=yb;ya++)
			*((DWORD*)bmp->ScanLine[ya]+xx)=0x00000000u;

		// generate text and its dimensions
		double xmk=xaxmark[i];
		AnsiString str;
		str.sprintf("%g",xmk);
		int wd=can->TextWidth(str);
		int wd2=wd/2;
		int he=can->TextHeight(str);

		xx-=wd2;
		if(xx<0) xx=0; else if(xx+wd+awd+1>=xs) continue;
		if(max>=-min)
		{
			ya=zy+SGR_XAX_LX;
			if(ya+he>=ys) ya=ys-he-1;
		}
		else
		{
			ya=zy-SGR_XAX_LX-he;
			if(ya<0) ya=0;
		}
		can->TextOutA(xx,ya,str);
	}

	// draw axis label
	can->Brush->Style=bsClear;
	can->TextOutA(xs-awd-1,zy+((max>=-min)?SGR_ALBL_DST:(-SGR_ALBL_DST-ahe)),"t [s]");

	// draw marks
	awd=can->TextWidth("y [Hz]");
	ahe=can->TextHeight("y [Hz]");
	for(i=0;i<yamn;i++)
	{
		int ya=zy-(int)(ss*yaxmark[i]);
		if(ya<=ahe) continue; else if(ya>=ys) ya=ys-1;

		can->MoveTo(SGR_YAX_X-SGR_YAX_MX,ya);
		can->LineTo(SGR_YAX_X,ya);

		// generate text and its dimensions
		double ymk=yaxmark[i];
		AnsiString str;
		str.sprintf("%g",ymk);
		int wd=can->TextWidth(str);
		int he=can->TextHeight(str);
		int he2=he/2;

		int xx=SGR_YAX_X-SGR_YAX_LX-wd;
		if(xx<0) xx=0;
		int yy=ya-he2;
		if(yy<ahe && ymk>0) continue; else if(yy>=(ys-he)) yy=ys-he-1;

		can->TextOutA(xx,yy,str);
	}

	// draw axis label
	can->Brush->Style=bsClear;
	can->TextOutA(SGR_YAX_X-awd-SGR_ALBL_DST,1,"y [Hz]");

	// draw axis direction
	TPoint tri[3];
	tri[0].x=SGR_YAX_X;
	tri[0].y=1;
	tri[1].x=SGR_YAX_X+SGR_TRI/2;
	tri[1].y=SGR_TRI+1;
	tri[2].x=SGR_YAX_X-SGR_TRI/2;
	tri[2].y=SGR_TRI+1;
	can->Pen->Style=psSolid;
	can->Brush->Color=clBlack;
	can->Polygon(tri,2);
	tri[0].x=xs-2;
	tri[0].y=zy;
	tri[1].x=xs-SGR_TRI-2;
	tri[1].y=zy+SGR_TRI/2;
	tri[2].x=xs-SGR_TRI-2;
	tri[2].y=zy-SGR_TRI/2;
	can->Polygon(tri,2);
  
	// draw waveform
	if(n>1)
	{
		// final y line first
		if(CBstep->Checked)
		{
			int ye=(int)((double)zy-(ss*gain));
			if(ye<0) ye=0; else if(ye>=ys) ye=ys-1;
			can->Pen->Color=clGray;
			can->Pen->Style=psDot;
			can->Brush->Style=bsClear;
			can->MoveTo(SGR_YAX_X,ye);
			can->LineTo(xs-1,ye);
		}

		// mow the waveform
		x=SGR_YAX_X;
		double xstep=(double)(xs-SGR_YAX_X-1)/(double)(n-1);
		for(i=0;i<n-1;i++)
		{
			// get y
			double ya=ss*y[i];
			double yb=ss*y[i+1];

			// get x
			int xa=(int)x;
			x+=xstep;
			int xb=(int)x;
			if(xa>=xs) xa=xs-1;
			if(xb>=xs) xb=xs-1;

			// limit y
			int yai=(int)((double)zy-ya);
			int ybi=(int)((double)zy-yb);
			if(yai<0) yai=0; else if(yai>=ys) yai=ys-1;
			if(ybi<0) ybi=0; else if(ybi>=ys) ybi=ys-1;

			// horizontal bit
			DWORD *pl=((DWORD*)bmp->ScanLine[yai])+xa;
			for(xa;xa<=xb;xa++)
				*pl++=0x000000FFu;

			// vertical bit
			if(yai>ybi)
			{
				int tmp=yai;
				yai=ybi;
				ybi=tmp;
			}
			for(yai;yai<=ybi;yai++)
				*((DWORD*)bmp->ScanLine[yai]+xb)=0x000000FFu;
		}
	}
}




void __fastcall TMSForm::imgDblClick(TObject *Sender)
{
	if(OD1->Execute())
	{
		if(strcmpi(ExtractFileExt(OD1->FileName).c_str(),".bmp")!=0)
			OD1->FileName=OD1->FileName+".bmp";
		img->Picture->Bitmap->PixelFormat=pf24bit;
		img->Picture->Bitmap->SaveToFile(OD1->FileName);
	}
}
//---------------------------------------------------------------------------





