//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - Rotary encoder decoder module
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#pragma hdrstop

#include "encoder.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

// encoder struct
TEnc enc;

//---------------------------------------------------------------------------
// Encoder - initialize
//---------------------------------------------------------------------------
int enc_init(int mode,int ppr,int treshold)
{
	if(mode!=ENC_SINGLE && mode!=ENC_QUAD)
		return(1);

	// store mode
	enc.mode=mode;

	// store pulse per revolition
	enc.ppr=ppr;

	// store treshold value
	enc.ofs=treshold;

	// init variables
	for(int i=0;i<4;i++)
	{
		enc.edge[i].last=1;
		enc.edge[i].first=1;
		enc.edge[i].cnt=0;;
		enc.edge[i].freq=0;
	}
	enc.dir=0;
	enc.freq=0;
	enc.rps=0;

	return(0);
}

//---------------------------------------------------------------------------
// Encoder - set pulse-per-revolution value
//---------------------------------------------------------------------------
int enc_set_ppr(int ppr)
{
	if(ppr<1 || ppr>1024)
		return(1);

	enc.ppr=ppr;

	return(0);
}

//---------------------------------------------------------------------------
// Encoder - decode 8-bit stereo unsigned stream
//---------------------------------------------------------------------------
int enc_decode(char *data,int count,int srate)
{
	int i,j,k;

	if(!data)
		return(1);

	// init variables
	int lcnt[4];
	for(i=0;i<4;i++)
	{
		// remember last pulse count
		lcnt[i]=enc.edge[i].cnt;
		// clear it
		enc.edge[i].cnt=0;
	}

	// no direction
	int dir=0;

	// ====== for each stereo sample ======
	int min[2]={256,256};
	int max[2]={-256,-256};
	for(i=0;i<count;i++)
	{
		// ------ for each channel ------
		for(j=0;j<enc.mode;j++)
		{
			// get current and previous sample, convert to signed
			int sa=((int)((i==0)?enc.smpl[j]:((uchar)data[(i-1)*2+j])))-0x80;
			int sb=((int)((uchar)data[i*2+j]))-0x80;

			// detect levels
			if(sa>max[j]) max[j]=sa;
			if(sa<min[j]) min[j]=sa;

			// --- for each edge ---
			for(k=0;k<2;k++)
			{
				TEncEdge *edge=&enc.edge[j*2+k];

				// invert sample for falling edge
				if(k)
				{
					sa=-sa;
					sb=-sb;
				}

				// detect zero cross
				if(sa<=enc.ofs && sb>=enc.ofs)
				{
					// rising edge detected - interpolate exact cross position
					edge->last=(double)i + (sb!=sa)?(double)(enc.ofs-sa)/(double)(sb-sa):0;
					edge->cnt++;

					// check direction in quadrature mode
					if(enc.mode==ENC_QUAD && j==0 && k==0)
						((int)((uchar)data[i*2+1])>0x80)?dir++:dir--;
				}
			}
		}
	}

	// store signal elvels
	for(i=0;i<2;i++)
		enc.lev[i]=(max[i]>-min[i])?max[i]:-min[i];

	// remember last sample from buffer
	for(i=0;i<2;i++)
		enc.smpl[i]=(uchar)data[(count-1)*2+i];


	// --- for each encoder channel ---
	double freq=0;
	int vcnt=0;
	for(i=0;i<enc.mode*2;i++)
	{
		TEncEdge *edge=&enc.edge[i];

		if(lcnt[i] && edge->cnt)
		{
			// valid - get new freq
			edge->freq=(double)srate*(double)edge->cnt/(edge->last-edge->first);
		}
		else
		{
			// invalid samples block
			edge->freq=0;
		}
		edge->first=edge->last-count;

		// sum valid results
		if(edge->freq)
		{
			freq+=edge->freq;
			vcnt++;
		}
	}

	// --- average freq ---
	if(vcnt)
		freq/=(double)vcnt;

	// apply direction
	if(dir<0)
		freq=-freq;

	// store direction
	enc.dir=(dir>=0)?1:-1;

	// store pulse freq
	enc.freq=freq;

	// store rev.per.sec.
	enc.rps=freq/(double)enc.ppr;

	return(0);
}

//---------------------------------------------------------------------------
// Encoder - get encoder structure
//---------------------------------------------------------------------------
int enc_get(TEnc **penc)
{
	if(!penc)
		return(1);

	// return pointer to encoder data
	*penc=&enc;

	return(0);
}

