//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - nasty graphs rendering functions
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------


#include <vcl.h>
#include <math.h>
#include <stdlib.h>

#pragma hdrstop

#include "graph.h"
#include "sb_inout.h"
#include "sys_model.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

TInData indat;

//---------------------------------------------------------------------------
// Initialize input graph data buffer
//---------------------------------------------------------------------------
int enc_graph_init_buffer(void)
{
	// get input buffer length
	int len;
	sb_get_input_len(&len);

	// allocate two buffers
	indat.data[0]=(char*)malloc(len*2);
	if(!indat.data[0])
	{
		// allocation error
		indat.data[0]=NULL;
		indat.data[1]=NULL;
		return(1);
	}
	memset((void*)indat.data[0],0x80,len*2);

	// set poitner for secondary buffer
	indat.data[1]=indat.data[0]+len;

	// no data ready
	indat.ready=0;

	// none used
	indat.used=0;

	return(0);
}

//---------------------------------------------------------------------------
// Close input graph buffer
//---------------------------------------------------------------------------
int enc_graph_close_buffer(void)
{
	if(indat.data[0])
		free((void*)indat.data[0]);
	indat.data[0]=NULL;

	return(0);
}

//---------------------------------------------------------------------------
// Store new input data into graph buffer
//---------------------------------------------------------------------------
int enc_graph_fill_buffer(char *data,int count)
{
	int next;

	// mark possibly ready buffer as not ready
	indat.ready=0;
	// check which buffer is NOT in use
	if(indat.used==0 || indat.used==2)
		next=0;
	else
		next=1;

	// leave if no buffer allocated
	if(!indat.data[next])
		return(1);

	// copy data into buffer
	memcpy((void*)indat.data[next],(void*)data,count*2);
	indat.len[next]=count;

	//select it as ready to use
	indat.ready=next+1;

	return(0);
}

//---------------------------------------------------------------------------
// Redraw input graph from buffer if something new there
//---------------------------------------------------------------------------
int enc_graph_draw_wave(TImage *img,int scale)
{
	int i,j,k;

	if(!img)
		return(1);

	// get scale step
	static int stbl[4]={10,5,2,1};
	if(scale<0)
		scale=3;
	int ss=stbl[scale];

	// check if there are data to render
	indat.used=indat.ready;

	// leave if nothing to draw
	if(!indat.used || !indat.data[indat.used-1])
		return(-1);

	// some data to draw there
	char *data=indat.data[indat.used-1];
	int count=indat.len[indat.used-1];

	// config bmp
	Graphics::TBitmap *bmp=img->Picture->Bitmap;
	bmp->Width=img->Width;
	bmp->Height=img->Height;
	bmp->PixelFormat=pf32bit;

	// get size
	int xs=bmp->Width;
	int ys=bmp->Height;

	// draw horizontal grid
	double gstep=(double)GR_GRIDN/ys;
	double gpos=-1;
	for(i=0;i<ys;i++)
	{
		DWORD *pl=(DWORD*)bmp->ScanLine[i];

		gpos+=gstep;
		if(gpos>=0)
		{
			for(j=0;j<xs;j++)
				*pl++=GR_GRID;
			gpos-=1;
		}
		else
		{
			memset((void*)pl,0x00,sizeof(DWORD)*xs);
		}
	}

	// vertical grid
	gpos=0 - ((double)xs/2-floor((double)xs/2*gstep)/gstep)*gstep;
	for(i=0;i<xs;i++)
	{
		gpos+=gstep;
		if(gpos>=0)
		{
			for(j=0;j<ys;j++)
			{
				DWORD *pl=(DWORD*)bmp->ScanLine[j]+i;
				*pl=GR_GRID;
			}
			gpos-=1;
		}
	}

	// find center rising edge
	int c;
	for(c=count/2;c<count-1;c++)
	{
		int sa=(int)((uchar)data[c*2+0])-0x80;
		int sb=(int)((uchar)data[(c+1)*2+0])-0x80;

		if(sa<=0 && sb>=0)
			break;
	}
	if(c==count-1)
		c=count/2;


	// now draw waveform
	for(k=0;k<2;k++)
	{
		// data limits
		uchar *wave_s=(uchar*)&data[k+0];
		uchar *wave_e=(uchar*)&data[k+(count-2)*2];

		// left/right part of graph
		for(i=-1;i<=1;i+=2)
		{
			uchar *wave=(uchar*)&data[k+(c-ss*(i<1))*2];
			int xpos=xs/2 - (i<0);
			do{
				// get neigboring samples A, B
				int sa=(int)wave[0]-0x80;
				int sb=(int)wave[ss<<1]-0x80;

				// rescale A, B to ysize
				int ya=sa*ys/258 + ys/2;
				int yb=sb*ys/258 + ys/2;
				// A<=B
				if(ya>yb) {int tmp=ya;ya=yb;yb=tmp;};

				// draw vertical line from A to B
				for(ya;ya<=yb;ya++)
				{
					DWORD *pl=(DWORD*)bmp->ScanLine[ya]+xpos;
					*pl=(k==0)?GR_TRACEA:GR_TRACEB;
				}

				// move to next sample
				wave+=(i*ss)<<1;
				xpos+=i;
			}while(xpos>=0 && xpos<xs && wave>=wave_s && wave<wave_e);
		}
	}

	// mark buffer as unused
	indat.used=0;

	// repaint
	img->Repaint();

	return(0);
}














//---------------------------------------------------------------------------
// Speed graph - init buffer
//---------------------------------------------------------------------------
int reg_graph_init(TSpdGraph *buf)
{
	int i,j;

	if(!buf)
		return(1);

	// clear pointers
	for(i=0;i<GR_MAX_TRACE;i++)
	{
		buf->traces[i]=NULL;
		buf->pos[i]=0;
	}

	// allocate initial trace buffers
	for(i=0;i<GR_MAX_TRACE;i++)
	{
		buf->traces[i]=(double*)malloc(sizeof(double)*1024);
		if(!buf->traces[i])
			return(2);

		// clear trace
		for(j=0;j<1024;j++)
			buf->traces[i][j]=GR_NO_TRACE;
	}


	buf->length=1024;
	buf->min=0;
	buf->max=100;
	buf->rate=1;

	return(0);
}

//---------------------------------------------------------------------------
// Speed graph - close buffer
//---------------------------------------------------------------------------
int reg_graph_close(TSpdGraph *buf)
{
	if(!buf)
		return(1);

	// loose buffers
	for(int i=0;i<GR_MAX_TRACE;i++)
	{
		free((void*)buf->traces[i]);
		buf->traces[i]=NULL;
	}
	buf->length=0;

	return(0);
}

//---------------------------------------------------------------------------
// Speed graph - close buffer
//---------------------------------------------------------------------------
int reg_graph_config(TSpdGraph *buf,double max,double sym,int rate)
{
	if(!buf)
		return(1);

	buf->max=max;
	buf->min=(!sym)?0:-max;

	if(rate<0)
		rate=0;
	buf->rate=rate+1;

	return(0);
}

//---------------------------------------------------------------------------
// Speed graph - add new sample into buffer
//---------------------------------------------------------------------------
int reg_graph_add(TImage *img,TSpdGraph *buf,double sample,int id)
{
	int i,j;

	if(!buf || !img || id>=GR_MAX_TRACE)
		return(1);

	// get graph width
	int xs=img->Width;

	// extend buffer?
	if(buf->length<xs)
	{
		// yes, it's too small for graph width
		for(i=0;i<GR_MAX_TRACE;i++)
		{
			// allocate new trace
			double *trace=(double*)malloc(sizeof(double)*xs);
			if(!trace)
				return(2);

			// copy original trace into new buffer
			memcpy((void*)&trace[0],(void*)&buf->traces[i][buf->pos[i]],sizeof(double)*(buf->length-buf->pos[i]));
			memcpy((void*)&trace[buf->length-buf->pos[i]],(void*)&buf->traces[i][0],sizeof(double)*buf->pos[i]);
			for(j=buf->length;j<xs;j++)
				trace[j]=GR_NO_TRACE;

			// replace buffers
			free((void*)buf->traces[i]);
			buf->traces[i]=trace;

			// set new position
			buf->pos[i]=buf->length;
		}

		// new length
		buf->length=xs;
	}

	// add new sample
	for(i=0;i<buf->rate;i++)
	{
		buf->traces[id][buf->pos[id]]=sample;
		buf->pos[id]++;
		if(buf->pos[id]>=buf->length)
			buf->pos[id]=0;
	}

	return(0);
}

//---------------------------------------------------------------------------
// Speed graph - Draw grid
//---------------------------------------------------------------------------
int reg_graph_grid(TImage *img)
{
	int i,j;

	if(!img)
		return(1);

	// config bmp
	Graphics::TBitmap *bmp=img->Picture->Bitmap;
	bmp->Width=img->Width;
	bmp->Height=img->Height;
	bmp->PixelFormat=pf32bit;

	// get size
	int xs=bmp->Width;
	int ys=bmp->Height;
  
	// draw horizontal grid
	double gstep=(double)GR_GRIDN/ys;
	double gpos=-1;
	for(i=0;i<ys;i++)
	{
		DWORD *pl=(DWORD*)bmp->ScanLine[i];

		gpos+=gstep;
		if(gpos>=0)
		{
			for(j=0;j<xs;j++)
				*pl++=GR_GRID;
			gpos-=1;
		}
		else
		{
			memset((void*)pl,0x00,sizeof(DWORD)*xs);
		}
	}

	// vertical grid
	gpos=-1;
	for(i=0;i<xs;i++)
	{
		gpos+=gstep;
		if(gpos>=0)
		{
			for(j=0;j<ys;j++)
			{
				DWORD *pl=(DWORD*)bmp->ScanLine[j]+i;
				*pl=GR_GRID;
			}
			gpos-=1;
		}
	}

	return(0);
}

//---------------------------------------------------------------------------
// Speed graph - draw single trace
//---------------------------------------------------------------------------
int reg_graph_draw_trace(TImage *img,TSpdGraph *buf,int id,DWORD color)
{
	if(!img || !buf || id>=GR_MAX_TRACE)
		return(1);

  // remember color
	buf->colors[id]=color;

	// config bmp
	Graphics::TBitmap *bmp=img->Picture->Bitmap;
	bmp->Width=img->Width;
	bmp->Height=img->Height;
	bmp->PixelFormat=pf32bit;

	// get size
	int xs=bmp->Width;
	int ys=bmp->Height;
	int ysd=(double)ys;

	// trace data
	double *data=&buf->traces[id][buf->pos[id]];
	double *beg=&buf->traces[id][0];
	double *end=&buf->traces[id][buf->length-1];
	int i;

	// draw trace
	for(i=xs-1;i>0;i--)
	{
		// get samples
		--data;
		if(data<beg)
			data=end;
		double sad=*data;

		double *datab=data;
		--datab;
		if(datab<beg)
			datab=end;
		double sbd=*datab;

		if(sad!=GR_NO_TRACE && sbd!=GR_NO_TRACE)
		{
			// rescale to graph height
			int sa=ys-(sad-buf->min)*ysd/(buf->max-buf->min+2)-1;
			// rescale to graph height
			int sb=ys-(sbd-buf->min)*ysd/(buf->max-buf->min+2)-1;

			// limit Y
			if(sa<0) sa=0; else if(sa>=ys) sa=ys-1;
			if(sb<0) sb=0; else if(sb>=ys) sb=ys-1;
			if(sa>sb) {int tmp=sa;sa=sb;sb=tmp;}

			for(sa;sa<=sb;sa++)
			{
				DWORD *pl=(DWORD*)bmp->ScanLine[sa];
				pl[i]=color;
			}
		}
	}

	return(0);
}

//---------------------------------------------------------------------------
// Speed graph - draw legend
//---------------------------------------------------------------------------
int reg_graph_draw_legend(TImage *img,TSpdGraph *buf,TFont *font,int n,char *str[])
{
	if(!img || !buf)
		return(1);

	if(n>GR_MAX_TRACE)
		n=GR_MAX_TRACE;

	// config bmp
	Graphics::TBitmap *bmp=img->Picture->Bitmap;
	bmp->Width=img->Width;
	bmp->Height=img->Height;
	bmp->PixelFormat=pf32bit;

	// get size
	int xs=bmp->Width;

	#define GR_LSEP 16
	#define GR_LSYM "---"

	int xpos=xs-GR_LSEP/2;
	for(int i=0;i<n;i++)
	{
		xpos-=bmp->Canvas->TextWidth(GR_LSYM+AnsiString(str[i]));
		DWORD clr=(buf->colors[i]>>16)|((buf->colors[i]<<16)&0x00FF0000u)|(buf->colors[i]&0x0000FF00u);
		bmp->Canvas->Font=font;
		bmp->Canvas->Font->Color=(TColor)clr;
		bmp->Canvas->Brush->Style=bsClear;
		bmp->Canvas->TextOutA(xpos,0,GR_LSYM+AnsiString(str[i]));
		xpos-=GR_LSEP;
	}

	return(0);
}











//---------------------------------------------------------------------------
// Pole/Zero map - draw background
//---------------------------------------------------------------------------
void PZ_draw_back(TImage *img,double ss)
{
	// canvas
	TCanvas *can=img->Canvas;
	img->Picture->Bitmap->Width=img->Width;
	img->Picture->Bitmap->Height=img->Height;

	// window size
	int xs=img->Width;
	int ys=img->Height;
	int xs2=xs/2;
	int ys2=ys/2;

	// background
	can->Brush->Color=clWhite;
	can->Brush->Style=bsSolid;
	can->Pen->Color=clBlack;
	can->Pen->Style=psSolid;
	can->Pen->Width=1;
	can->Rectangle(0,0,xs,ys);
	can->Pen->Style=psDot;
	can->MoveTo(xs2,0);
	can->LineTo(xs2,ys-1);
	can->MoveTo(0,ys2);
	can->LineTo(xs-1,ys2);

	// draw axis direction
	TPoint tri[3];
	tri[0].x=xs2;
	tri[0].y=1;
	tri[1].x=xs2+PZMAP_TRI/2;
	tri[1].y=PZMAP_TRI+1;
	tri[2].x=xs2-PZMAP_TRI/2;
	tri[2].y=PZMAP_TRI+1;
	can->Pen->Style=psSolid;
	can->Brush->Color=clBlack;
	can->Polygon(tri,2);
	tri[0].x=xs-2;
	tri[0].y=ys2;
	tri[1].x=xs-PZMAP_TRI-2;
	tri[1].y=ys2+PZMAP_TRI/2;
	tri[2].x=xs-PZMAP_TRI-2;
	tri[2].y=ys2-PZMAP_TRI/2;
	can->Polygon(tri,2);

	// draw axis label
	can->Brush->Style=bsClear;
	int wd=can->TextWidth("Im");
	can->TextOutA(xs2-PZMAP_AXLBL_DIST-wd,1,"Im");
	wd=can->TextWidth("Re");
	can->TextOutA(xs-1-wd-1,ys2+PZMAP_AXLBL_DIST,"Re");

	// draw critical circle
	can->Brush->Style=bsClear;
	can->Pen->Color=clGray;
	can->Pen->Style=psDash;
	can->Ellipse(xs2+(int)ss,ys2+(int)ss,xs2-(int)ss,ys2-(int)ss);

	// draw scale labels
	wd=can->TextWidth("1");
	int he=can->TextHeight("1");
	can->TextOutA(xs2+3,ys2-(int)ss-he-1,"1");
	can->TextOutA(xs2+(int)ss+3,ys2-he-1,"1");
}

//---------------------------------------------------------------------------
// Pole/Zero map - draw poles/zeros
//---------------------------------------------------------------------------
void PZ_draw_pz(TImage *img,TPoZe *pz,int n,int type,double ss)
{
	// canvas
	TCanvas *can=img->Canvas;

	// window size
	int xs=img->Width;
	int ys=img->Height;
	int xs2=xs/2;
	int ys2=ys/2;

	// and the poles
	can->Pen->Color=(type==PZMAP_TYPE_ZERO)?clBlue:clRed;
	for(int i=0;i<n;i++)
	{
		// get coordinates
		double x=pz[i].re;
		double y=pz[i].im;

		// limit max size just in case
		if(fabs(x)>100 || fabs(y)>100)
			continue;

		// draw crosses
		can->Pen->Width=2;
		can->Pen->Style=psSolid;
		if(type==PZMAP_TYPE_ZERO)
		{
			// zero
			can->Ellipse(xs2+(int)(ss*x-PZMAP_PD),ys2+(int)(ss*y-PZMAP_PD),xs2+(int)(ss*x+PZMAP_PD),ys2+(int)(ss*y+PZMAP_PD));
		}
		else
		{
			// pole
			can->MoveTo(xs2+(int)(ss*x-PZMAP_PD),ys2+(int)(ss*y-PZMAP_PD));
			can->LineTo(xs2+(int)(ss*x+PZMAP_PD),ys2+(int)(ss*y+PZMAP_PD));
			can->MoveTo(xs2+(int)(ss*x+PZMAP_PD),ys2+(int)(ss*y-PZMAP_PD));
			can->LineTo(xs2+(int)(ss*x-PZMAP_PD),ys2+(int)(ss*y+PZMAP_PD));
		}

		if(y!=0)
		{
			can->Pen->Width=1;
			can->Pen->Style=psDot;
			can->MoveTo(xs2+(int)(ss*x),ys2+(int)(ss*y));
			can->LineTo(xs2+0,ys2+0);
		}
	}
}

