//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - VCL GUI module
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <vcl.h>
#include <windows.h>
#include <mmsystem.h>
#include <inifiles.hpp>

#pragma hdrstop

#include "main.h"
#include "PZ_form.h"
#include "MPZ_form.h"
#include "MS_form.h"
#include "AB_form.h"
#include "graph.h"
#include "sb_inout.h"
#include "encoder.h"
#include "regulator.h"
#include "sys_model.h"
#include "pwm.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMForm *MForm;

// stream recorder
TRec rec;


//---------------------------------------------------------------------------
// SB input - user callback function
//---------------------------------------------------------------------------
void sb_input_user_callback(UINT uMsg,char *data,int count,int srate)
{
	if(uMsg==WIM_DATA)
	{
		// === INPUT BUFFER READY ===

		// --- process input data ---
		// decode encoder waveform
		enc_decode(data,count,srate);

		// get encoder output
		TEnc *enc;
		enc_get(&enc);


		// --- update regulator ---
		// calculate regulator output (one sample)
		double duty;
		reg_update(enc->rps,&duty);

		// rescale -1 to +1 into 0 to 1 for symetrical output
		double duty_out=duty;
		if(!reg_get_man() && reg_get_limit())
		{
			duty*=0.5;
			duty_out=duty+0.5;
		}

		// set new PWM duty cycle
		pwm_set_duty(duty_out);



		// --- online system identification ---
		// identify system
		double y_pred;
		sys_idn_update(duty,enc->rps,&y_pred,NULL,NULL);



		// --- update regulator parameters in adaptive mode ---
		if(sys_idn_adaptive())
		{
			// get Ziegler-Nichols parameters
			double k,ti,td;
			sys_idn_get_zn(&k,&ti,&td);

			// and set them to the regulator
			reg_set_PID(k,ti,td);
		}


		// --- store into record ---
		// get current desired value
		double w;
		reg_get_w(&w);

		if(rec.op==RECOP_NEW)
		{
			// new file
			rec.op=RECOP_ADD;
			rec.time=0.0;
			if(rec.fw)
			{
				fclose(rec.fw);
				rec.fw=NULL;
			}
			rec.fw=fopen(rec.fname,"wt");
			if(rec.fw)
				fprintf(rec.fw,"t;duty;w;y;y_pred\ns;%%;Hz;Hz;Hz\n");
		}
		else if(rec.fw && rec.op==RECOP_CLOSE)
		{
			rec.op=0;
			fclose(rec.fw);
			rec.fw=NULL;
		}
		if(rec.fw && rec.op==RECOP_ADD)
		{
			// add data row
			fprintf(rec.fw,"%f;%f;%f;%f;%f\n",rec.time,duty,w,enc->rps,y_pred);

			// update time
			double Ts;
			reg_get_T(&Ts);
			rec.time+=Ts;
		}



		// --- add recieved waveform into scope buffer ---
		enc_graph_fill_buffer(data,count);
	}

	// send refresh message to VCL GUI (asynchronously - it runs in different thread)
	PostMessage(MForm->Handle,CM_INPUT_REFRESH,uMsg,0);
}


//---------------------------------------------------------------------------
// SB output - user callback function
//---------------------------------------------------------------------------
void sb_out_user_callback(UINT uMsg,char *data,int count,int srate)
{
	if(uMsg==WOM_DONE)
	{
		// === OUPUT BUFFER DONE (time critical!) ===

		// fill output buffer with PWM waveform at current duty cycle
		pwm_gen(data,count,srate);
	}
}




//---------------------------------------------------------------------------
// Form initializer
//---------------------------------------------------------------------------
__fastcall TMForm::TMForm(TComponent* Owner)
	: TForm(Owner)
{
}


//---------------------------------------------------------------------------
// Form show
//---------------------------------------------------------------------------
void __fastcall TMForm::FormShow(TObject *Sender)
{
	// paths
	apppath=GetCurrentDir();
	helppath=apppath+"\\"+HELPFILE;

	// load ini config
	if(FileExists(apppath+"\\"+INIFILE))
		cini=new TIniFile(apppath+"\\"+INIFILE);
	else
		cini=NULL;
	LoadIni(cini);

	// double buffered graphs components
	GBin->DoubleBuffered=true;
	GBspd->DoubleBuffered=true;
	GBduty->DoubleBuffered=true;

	// initialize graphs
	reg_graph_init(&spd);
	reg_graph_config(&spd,UDspdMax->Position,CBspdSym->Checked,CBspdRate->ItemIndex);
	reg_graph_init(&duty);
	reg_graph_config(&duty,100,0,CBdutyRate->ItemIndex);

	// init regulator
	reg_init(UDper->Position);

	// initial pulser period
	EregPperChange(NULL);

	// initialize PWM output
	pwm_init();
	EpfreqChange(NULL);

	// initialize system identificator
	sys_idn_init();
	EarxFpChange(NULL);
	set_manual_PID();
	set_auto_zn_PID();

	// init recorder
	rec.op=0;
	rec.fw=NULL;
	strcpy(rec.fname,(apppath+"\\"+RECFILE).c_str());

	// === MAKE LIST OF AUDIO DEVICES ===
	// get in/out device counts
	int cin,cout;
	sb_inout_dev_count(&cin,&cout);

	// make list of input devices
	CBinDevs->Clear();
	for(int i=0;i<cin;i++)
	{
		char name[32];
		sb_get_output_dev(0,i,name);
		CBinDevs->Items->Add(AnsiString(name));
	}
	CBinDevs->ItemIndex=0;

	// make list of output devices
	CBoutDevs->Clear();
	for(int i=0;i<cout;i++)
	{
		char name[32];
		sb_get_output_dev(1,i,name);
		CBoutDevs->Items->Add(AnsiString(name));
	}
	CBoutDevs->ItemIndex=0;
}


//---------------------------------------------------------------------------
// Form close
//---------------------------------------------------------------------------
void __fastcall TMForm::FormClose(TObject *Sender, TCloseAction &Action)
{
	// cleanup
	sb_close();
	enc_graph_close_buffer();
	reg_graph_close(&spd);
	reg_graph_close(&duty);
	sys_idn_close();

	// close recorder
	if(rec.fw)
	{
		fclose(rec.fw);
		rec.fw=NULL;
	}

	// save config.ini
	SaveIni(cini);

  // close config.ini
	if(cini)
		delete(cini);
}

//---------------------------------------------------------------------------
// Start/stop button
//---------------------------------------------------------------------------
void __fastcall TMForm::BtnGoClick(TObject *Sender)
{
	static srlst[]={8000,11025,16000,22050,32000,44100,48000,96000,192000};
	static run=0;


	if(!run)
	{
		// --- init encoder ---
		enc_init(ENC_QUAD,UDppr->Position,16);


		// --- init PWM ---
		pwm_set_invert(CBinv->Checked);


		// --- init regulator ---
		// regulator init
		reg_init(UDper->Position);
		// manual mode?
		if(RBman->Checked)
			RBmanClick(NULL);
		// set it with GUI parameters
		set_manual_PID();
		// symetrical limitation?
		CBspdSymClick(NULL);


		// --- config system identificator ---
		// basic config (samping, patterns count)
		double Ts;
		reg_get_T(&Ts);
		sys_idn_config(UDarxSn->Position,Ts);
		// set limits (iteration cycles and error)
		set_idn_limits();
		// set desired mode
		sys_idn_adaptive(BtnAdaptive->Down);


		// --- init recorder ---
		CBrecClick(NULL);


		// --- try to init SB ---
		if(!sb_init(CBinDevs->ItemIndex,CBoutDevs->ItemIndex,srlst[CBsrate->ItemIndex],UDper->Position,UDodel->Position,sb_input_user_callback,sb_out_user_callback))
		{
			// success
			run^=1;

			// init input wave scope buffer
			enc_graph_init_buffer();
		}
		else
		{
			SB->Panels->Items[1]->Text=" SB opening failed!";
		}
	}
	else
	{
		// --- close SB ---
		sb_close();
		run^=1;
	}

	BtnGo->Caption=(!run)?"Start":"Stop";
}


//---------------------------------------------------------------------------
// Input refresh request message handler,
// redraws input graph, prints some infos
//---------------------------------------------------------------------------
void TMForm::InputRefresh(TMessage msg)
{
	if(msg.WParam==WIM_DATA)
	{
		// === NEW DATA BLOCK DONE ===
		// draw graph
		int ret=enc_graph_draw_wave(MForm->ImgEnc,CBscale->ItemIndex);

		// redraw status
		if(ret==0)
		{
			// get encoder structure
			TEnc *enc;
			enc_get(&enc);

			// print encoder data
			Efreq->Text=FormatFloat("0.0 Hz",enc->rps);
			Erpm->Text=FormatFloat("0 rpm",60.0*enc->rps);

			// get duty cycle
			double pdut;
			pwm_get_duty(&pdut);

			// rescale -1 to +1 into 0 to 1 for symetrical output
			if(!reg_get_man() && reg_get_limit())
				pdut=0.5*(pdut+1);

			// get regualator desired value
			double w;
			reg_get_w(&w);

			// regulator limitation state
			int lim=reg_get_x_limit();
			Exlim->Color=(lim)?clRed:clBtnFace;


			// identify system
			double th[SYS_N]={0};
			double kp[2]={0,0};
			double yp=0.0;
			double err=0.0;
			int cyc=0;
			sys_idn_get_model(&yp,th,kp,&err,&cyc);

			double merr=0.0;
			int mcyc=0;
			sys_idn_limit(&mcyc,&merr);

			// get Ziegler-Nichols parameters
			double k,ti,td;
			sys_idn_get_zn(&k,&ti,&td);

			// config regulator if adaptive mode enabled
			if(sys_idn_adaptive())
			{
				print_PID_config(k,ti,td);
			}

			// print ARX model
			TStaticText *arx[]={STarxB1,STarxB2,STarxB3,STarxA1,STarxA2,STarxA3};
			for(int i=0;i<6;i++)
				arx[i]->Caption=FormatFloat("0.0000",th[i]);

			// and critical parameters
			EarxKk->Text=FormatFloat("0.000",kp[0]);
			EarxTk->Text=FormatFloat("0.000",kp[1]);

			// and PID recommendation
			EarxTi->Text=FormatFloat("0.000",ti);
			EarxTd->Text=FormatFloat("0.000",td);
			EarxK->Text=FormatFloat("0.000",k);

			// and status
			SB->Panels->Items[3]->Text=" SD cycles: " + IntToStr(cyc) + "/" + IntToStr(mcyc);
			SB->Panels->Items[4]->Text=" SD error: " + FormatFloat("0.000",err) + "/" + FormatFloat("0.000",merr);

			// draw speed graph
			reg_graph_add(ImgSpd,&spd,enc->rps,0);
			reg_graph_add(ImgSpd,&spd,w,1);
			reg_graph_add(ImgSpd,&spd,yp,2);
			reg_graph_grid(ImgSpd);
			reg_graph_draw_trace(ImgSpd,&spd,1,GR_TRACEB);
			if(BtnArxTrace->Down)
				reg_graph_draw_trace(ImgSpd,&spd,2,GR_TRACED);
			reg_graph_draw_trace(ImgSpd,&spd,0,GR_TRACEA);
			char *leg[]={"y","w","ym"};
			reg_graph_draw_legend(ImgSpd,&spd,LegendFont->Font,3,leg);
			ImgSpd->Repaint();

			// draw duty graph
			reg_graph_add(ImgDuty,&duty,100*pdut,0);
			reg_graph_grid(ImgDuty);
			reg_graph_draw_trace(ImgDuty,&duty,0,GR_TRACEC);
			ImgDuty->Repaint();
			Eduty->Text=IntToStr((int)(100*pdut));
			if(!RBman->Checked)
				TBpwm->Position=100*(1.0-pdut);
			else
				SBtest->Position=enc->rps;

			// draw critical poles if enabled
			if(PZForm->Visible)
				PZForm->DrawPZ();

			// draw model poles/zeros if enabled
			if(MPZForm->Visible)
				MPZForm->DrawPZ();

			// draw step response
			if(MSForm->Visible)
				MSForm->DrawStep();
		}
	}
	else if(msg.WParam==WIM_OPEN)
	{
		// === INPUT OPEN ===
		SB->Panels->Items[1]->Text=" SB opened ...";
	}
	else if(msg.WParam==WIM_CLOSE)
	{
		// === INPUT CLOSE ===
		SB->Panels->Items[1]->Text=" SB closed ...";
	}
}




//---------------------------------------------------------------------------
// Desired value auto-pulser
//---------------------------------------------------------------------------
void __fastcall TMForm::TimerPulserTimer(TObject *Sender)
{
	static int state=0;

	// center 'w' position
	int w=SBtest->Position;
	int w_pkpk=UdregPpkpk->Position/2;

	if(wp_mode==WP_PULSE)
	{
		// === pulse ===
		if(state)
			w+=w_pkpk;
		else
			w-=w_pkpk;
	}
	else if(wp_mode==WP_RAND)
	{
		// === random ===
		w+=(2*w_pkpk*rand())/RAND_MAX - w_pkpk;
	}
	else if(wp_mode==WP_STEP)
	{
		// === steper ===
		int wd;
		if(wp_step<=WP_STEPN)
			wd=(2*w_pkpk*wp_step)/WP_STEPN;
		else
			wd=(2*w_pkpk*(WP_STEPN*2-wp_step))/WP_STEPN;

		w+=(wd-w_pkpk);

		wp_step++;
		if(wp_step>=WP_STEPN*2)
    	wp_step=0;
	}

	// limit amplitude
	if(w>UDspdMax->Position)
		w=UDspdMax->Position;
	else if(w<0)
		w=0;

	// set 'w'
	reg_set_w(w);

	// negate wave polarity
	state^=1;
}







//---------------------------------------------------------------------------
void __fastcall TMForm::EspdMaxChange(TObject *Sender)
{
	SBtest->Min=(CBspdSym->Checked)?-UDspdMax->Position:0;
	SBtest->Max=UDspdMax->Position;

	reg_graph_config(&spd,UDspdMax->Position,CBspdSym->Checked,CBspdRate->ItemIndex);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::CBspdSymClick(TObject *Sender)
{
	reg_graph_config(&spd,UDspdMax->Position,CBspdSym->Checked,CBspdRate->ItemIndex);
	reg_set_limit(CBspdSym->Checked);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::CBspdRateChange(TObject *Sender)
{
	CBdutyRate->ItemIndex=CBspdRate->ItemIndex;
	reg_graph_config(&spd,UDspdMax->Position,CBspdSym->Checked,CBspdRate->ItemIndex);
	reg_graph_config(&duty,100,0,CBspdRate->ItemIndex);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::CBdutyRateChange(TObject *Sender)
{
	CBspdRate->ItemIndex=CBdutyRate->ItemIndex;
	reg_graph_config(&duty,100,0,CBspdRate->ItemIndex);
	reg_graph_config(&spd,UDspdMax->Position,CBspdSym->Checked,CBspdRate->ItemIndex);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::SBtestChange(TObject *Sender)
{
	reg_set_w(SBtest->Position);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EmanwKeyPress(TObject *Sender, char &Key)
{
	if(Key==VK_RETURN)
	{
		SBtest->Position=Udmanw->Position;
		reg_set_w(Udmanw->Position);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EpfreqChange(TObject *Sender)
{
	pwm_set_freq(UDfreq->Position);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EpprChange(TObject *Sender)
{
	enc_set_ppr(UDppr->Position);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::TBpwmChange(TObject *Sender)
{
	if(RBman->Checked)
		reg_set_man(1,0.01*(double)(100-TBpwm->Position));
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EnpwmKeyPress(TObject *Sender, char &Key)
{
	if(Key==VK_RETURN && RBman->Checked)
	{
		TBpwm->Position=UDnpwm->Position;
		reg_set_man(1,0.01*(double)(UDnpwm->Position));
	}
}
//---------------------------------------------------------------------------
void __fastcall TMForm::RBmanClick(TObject *Sender)
{
	TBpwmChange(NULL);
	TimerPulser->Enabled=false;
}
//---------------------------------------------------------------------------
void __fastcall TMForm::RBautoClick(TObject *Sender)
{
	reg_set_man(0,0);
	TimerPulser->Enabled=false;
	SBtestChange(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::RBautoPClick(TObject *Sender)
{
	reg_set_man(0,0);
	wp_mode=WP_PULSE;
	TimerPulser->Enabled=true;
	EregPperChange(NULL);
	SBtestChange(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::RBautoRClick(TObject *Sender)
{
	reg_set_man(0,0);
	wp_mode=WP_RAND;
	TimerPulser->Enabled=true;
	EregPperChange(NULL);
	SBtestChange(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::RBautoSClick(TObject *Sender)
{
	reg_set_man(0,0);
	wp_mode=WP_STEP;
	wp_step=0;
	TimerPulser->Enabled=true;
	EregPperChange(NULL);
	SBtestChange(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EregPperChange(TObject *Sender)
{
	if(wp_mode==WP_PULSE)
		TimerPulser->Interval=UDregPper->Position/2;
	else
		TimerPulser->Interval=UDregPper->Position;
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EarxFpChange(TObject *Sender)
{
	sys_idn_filter(UDarxFp->Position);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EarxSnKeyPress(TObject *Sender, char &Key)
{
	if(Key==VK_RETURN)
		sys_idn_config(UDarxSn->Position,0);
}
void __fastcall TMForm::EregSkKeyPress(TObject *Sender, char &Key)
{
	if(key_press_float(Key))
		set_manual_PID();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::EznCkKeyPress(TObject *Sender, char &Key)
{
	if(key_press_float(Key))
		set_auto_zn_PID();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::BtnAutoTuneClick(TObject *Sender)
{
	// get Ziegler-Nichols parameters
	double k,ti,td;
	sys_idn_get_zn(&k,&ti,&td);

	// config regulator
	reg_set_PID(k,ti,td);
	print_PID_config(k,ti,td);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::BtnArxClipClick(TObject *Sender)
{
	char str[1024];
	sys_idn_print_system(str,0);
	text2clip(str);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::BtnArxClipTfClick(TObject *Sender)
{
	char str[1024];
	sys_idn_print_system(str,1);
	text2clip(str);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::CBarxEnabClick(TObject *Sender)
{
	sys_idn_state(CBarxEnab->Checked);
}
//---------------------------------------------------------------------------
void TMForm::set_idn_limits(void)
{
	double nerr=edit_to_float(EarxLrnE,0.001,1000,1,"0.000");
	int ncyc=UDarxLrnC->Position;
	sys_idn_limit(ncyc,nerr);
}
void __fastcall TMForm::EarxLrnEKeyPress(TObject *Sender, char &Key)
{
	if(key_press_float(Key))
		set_idn_limits();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::BtnArxRstClick(TObject *Sender)
{
	sys_idn_reset(1);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::BtnAdaptiveClick(TObject *Sender)
{
	sys_idn_adaptive(BtnAdaptive->Down);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::MMexitClick(TObject *Sender)
{
	Close();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::MMaboutClick(TObject *Sender)
{
	AForm->ShowModal();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::MMhelpClick(TObject *Sender)
{
	if(helppath=="" || !FileExists(helppath))
		MessageBox(Handle,"Missing help file!","Error ...",MB_ICONEXCLAMATION);
	else
		ShellExecute(Handle,"open",helppath.c_str(),NULL,NULL,SW_SHOWDEFAULT);
}
//---------------------------------------------------------------------------
void __fastcall TMForm::CBrecClick(TObject *Sender)
{
	if(CBrec->Down)
		rec.op=RECOP_NEW;
	else if(rec.fw)
		rec.op=RECOP_CLOSE;
}
//---------------------------------------------------------------------------
void __fastcall TMForm::ImgEncDblClick(TObject *Sender)
{
	if(OD1->Execute())
	{
		if(strcmpi(ExtractFileExt(OD1->FileName).c_str(),".bmp")!=0)
			OD1->FileName=OD1->FileName+".bmp";
		ImgEnc->Picture->Bitmap->PixelFormat=pf24bit;
		ImgEnc->Picture->Bitmap->SaveToFile(OD1->FileName);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMForm::ImgSpdDblClick(TObject *Sender)
{
	if(OD1->Execute())
	{
		if(strcmpi(ExtractFileExt(OD1->FileName).c_str(),".bmp")!=0)
			OD1->FileName=OD1->FileName+".bmp";
		ImgSpd->Picture->Bitmap->PixelFormat=pf24bit;
		ImgSpd->Picture->Bitmap->SaveToFile(OD1->FileName);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMForm::ImgDutyDblClick(TObject *Sender)
{
	if(OD1->Execute())
	{
		if(strcmpi(ExtractFileExt(OD1->FileName).c_str(),".bmp")!=0)
			OD1->FileName=OD1->FileName+".bmp";
		ImgDuty->Picture->Bitmap->PixelFormat=pf24bit;
		ImgDuty->Picture->Bitmap->SaveToFile(OD1->FileName);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMForm::MMcpolesClick(TObject *Sender)
{
	PZForm->Show();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::MMmpzmapClick(TObject *Sender)
{
	MPZForm->Show();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::MMarxStepClick(TObject *Sender)
{
	MSForm->Show();
}
//---------------------------------------------------------------------------
void __fastcall TMForm::CBinvClick(TObject *Sender)
{
	pwm_set_invert(CBinv->Checked);
}
//---------------------------------------------------------------------------








//---------------------------------------------------------------------------
// set PID manually
//---------------------------------------------------------------------------
void TMForm::set_manual_PID(void)
{
	double k=edit_to_float(EregSk,0.0001,1000,1,"0.0000");
	double ti=edit_to_float(EregSti,0.0001,1000,1,"0.0000");
	double td=edit_to_float(EregStd,0.0001,1000,1,"0.0000");
	double N=edit_to_float(EregSn,0.1,50,3,"0.0");
	double beta=edit_to_float(EregSbeta,0.000,1.000,1.000,"0.000");
	double tt=edit_to_float(EregStt,0.001,1000,1.000,"0.000");
	reg_set_PID_ext(k,ti,td,N,beta,tt);
}
//---------------------------------------------------------------------------
// display PID settings
//---------------------------------------------------------------------------
void TMForm::print_PID_config(double k,double ti,double td)
{
	EregSk->Text=FormatFloat("0.0000",k);
	EregSti->Text=FormatFloat("0.0000",ti);
	EregStd->Text=FormatFloat("0.0000",td);
}

//---------------------------------------------------------------------------
// set Ziegler-Nichols multipliers for auto-tune
//---------------------------------------------------------------------------
void TMForm::set_auto_zn_PID(void)
{
	double k=edit_to_float(EznCk,0.001,10,1,"0.000");
	double ti=edit_to_float(EznCti,0.001,10,1,"0.000");
	double td=edit_to_float(EznCtd,0.001,10,1,"0.000");
	sys_idn_config_zn(k,ti,td);
}

//---------------------------------------------------------------------------
// edit fix double
//---------------------------------------------------------------------------
int TMForm::key_press_float(char &Key)
{
	if(Key==VK_RETURN)
	{
		Key=0x00;
		return(1);
	}
	else if(Key=='.' || Key==',')
		Key=DecimalSeparator;
	return(0);
}

//---------------------------------------------------------------------------
// edit text to double
//---------------------------------------------------------------------------
double TMForm::edit_to_float(TEdit *edit,double min,double max,double def,char *fmt)
{
	AnsiString str=edit->Text;
	for(int i=1;i<=str.Length();i++)
		if(str.operator[](i)==',' || str.operator[](i)=='.')
			str.operator[](i)=DecimalSeparator;

	double val=StrToFloatDef(str,def);
	if(val<min)
		val=min;
	else if(val>max)
		val=max;
	edit->Text=FormatFloat(fmt,val);
	return(val);
}

//---------------------------------------------------------------------------
// Copy string into clipboard
//---------------------------------------------------------------------------
int TMForm::text2clip(char *str)
{
	HGLOBAL cbmem;
	char *cstr;

	// open clipboard
	if(!OpenClipboard(NULL))
	{
		// opening clipboard failed
		return(1);
	}

	// empty cliboard
	EmptyClipboard();

	// allocate new memory
	int len=strlen(str);
	cbmem=GlobalAlloc(GMEM_MOVEABLE,len+1);
	if(cbmem==NULL)
	{
		// allocation failed
		CloseClipboard();
		return(2);
	}

	// copy text
	cstr=(char*)GlobalLock(cbmem);
	strcpy(cstr,str);
	cstr[len]='\0';
	GlobalUnlock(cbmem);

	// set it into cliboard
	SetClipboardData(CF_TEXT,cbmem);

	// close clipboard
	CloseClipboard();

	return(0);
}

//---------------------------------------------------------------------------
// config.ini - read integer to TUpDown
//---------------------------------------------------------------------------
int IniReadUpDown(TIniFile *ini,AnsiString sec,AnsiString key,int def,TUpDown *ud)
{
	int num=ini->ReadInteger(sec,key,def);
	if(num<ud->Min)
		num=ud->Min;
	else if(num>ud->Max)
		num=ud->Max;

	ud->Position=num;

	return(num);
}

//---------------------------------------------------------------------------
// config.ini - read float string to TEdit
//---------------------------------------------------------------------------
AnsiString IniReadEditFloat(TIniFile *ini,AnsiString sec,AnsiString key,double def,TEdit *edit)
{
	AnsiString str=ini->ReadString(sec,key,FloatToStr(def));

	for(int i=1;i<=str.Length();i++)
		if(str.operator[](i)==',' || str.operator[](i)=='.')
			str.operator[](i)=DecimalSeparator;

	edit->Text=str;

	return(str);
}

//---------------------------------------------------------------------------
// config.ini - read settings
//---------------------------------------------------------------------------
void TMForm::LoadIni(TIniFile *ini)
{
	int num;

	if(!ini)
		return;

	// path
	OD1->InitialDir=ini->ReadString("PATH","last","");

	// SB setup
	num=ini->ReadInteger("SB","sample_rate",7);
	if(num<0)
		num=0;
	else if(num>=CBsrate->Items->Count)
		num=CBsrate->Items->Count-1;
	CBsrate->ItemIndex=num;
	IniReadUpDown(ini,"SB","period",100,UDper);
	IniReadUpDown(ini,"SB","delay",20,UDodel);

	// encoder setup
	IniReadUpDown(ini,"ENCODER","pulse_per_rev",24,UDppr);

	// speed setup
	IniReadUpDown(ini,"SPEED","max",50,UDspdMax);
	CBspdSym->Checked=ini->ReadInteger("SPEED","sym",0)!=0;

	// PWM setup
	IniReadUpDown(ini,"PWM","freq",480,UDfreq);

	// regulator
	IniReadUpDown(ini,"REGULATOR","gen_period",4000,UDregPper);
	IniReadUpDown(ini,"REGULATOR","gen_pkpk",20,UdregPpkpk);
	IniReadEditFloat(ini,"REGULATOR","k",0.01,EregSk);
	IniReadEditFloat(ini,"REGULATOR","Ti",1,EregSti);
	IniReadEditFloat(ini,"REGULATOR","Td",0.001,EregStd);
	IniReadEditFloat(ini,"REGULATOR","Tt",1,EregStt);
	IniReadEditFloat(ini,"REGULATOR","N",3,EregSn);
	IniReadEditFloat(ini,"REGULATOR","beta",1,EregSbeta);

	// model
	CBarxEnab->Checked=ini->ReadInteger("MODEL","enabled",1)!=0;
	BtnArxTrace->Down=ini->ReadInteger("MODEL","show_pred",1)!=0;
	IniReadUpDown(ini,"MODEL","sd_patterns",200,UDarxSn);
	IniReadUpDown(ini,"MODEL","sd_cycles",50,UDarxLrnC);
	IniReadEditFloat(ini,"MODEL","sd_error",1,EarxLrnE);
	IniReadUpDown(ini,"MODEL","sd_filter",20,UDarxFp);

	// auto-tuning
	IniReadEditFloat(ini,"AUTO","zn_k",0.3,EznCk);
	IniReadEditFloat(ini,"AUTO","zn_Ti",1.0,EznCti);
	IniReadEditFloat(ini,"AUTO","zn_Td",0.125,EznCtd);

	// model step graph
	IniReadUpDown(ini,"STEP","max_time",10,MSForm->UDlen);
}

//---------------------------------------------------------------------------
// config.ini - write settings
//---------------------------------------------------------------------------
void TMForm::SaveIni(TIniFile *ini)
{
	if(!ini)
		return;

	// path
	if(OD1->FileName!="")
		ini->WriteString("PATH","last",ExtractFilePath(OD1->FileName));

	// SB setup
	ini->WriteInteger("SB","sample_rate",CBsrate->ItemIndex);
	ini->WriteInteger("SB","period",UDper->Position);
	ini->WriteInteger("SB","delay",UDodel->Position);

	// encoder setup
	ini->WriteInteger("ENCODER","pulse_per_rev",UDppr->Position);

	// speed setup
	ini->WriteInteger("SPEED","max",UDspdMax->Position);
	ini->WriteInteger("SPEED","sym",(CBspdSym->Checked)?1:0);

	// PWM setup
	ini->WriteInteger("PWM","freq",UDfreq->Position);

	// regulator
	ini->WriteInteger("REGULATOR","gen_period",UDregPper->Position);
	ini->WriteInteger("REGULATOR","gen_pkpk",UdregPpkpk->Position);
	ini->WriteString("REGULATOR","k",EregSk->Text);
	ini->WriteString("REGULATOR","Ti",EregSti->Text);
	ini->WriteString("REGULATOR","Td",EregStd->Text);
	ini->WriteString("REGULATOR","Tt",EregStt->Text);
	ini->WriteString("REGULATOR","N",EregSn->Text);
	ini->WriteString("REGULATOR","beta",EregSbeta->Text);

	// model
	ini->WriteInteger("MODEL","enabled",(CBarxEnab->Checked)?1:0);
	ini->WriteInteger("MODEL","show_pred",(BtnArxTrace->Down)?1:0);
	ini->WriteInteger("MODEL","sd_patterns",UDarxSn->Position);
	ini->WriteInteger("MODEL","sd_cycles",UDarxLrnC->Position);
	ini->WriteString("MODEL","sd_error",EarxLrnE->Text);
	ini->WriteInteger("MODEL","sd_filter",UDarxFp->Position);

	// auto-tuning
	ini->WriteString("AUTO","zn_k",EznCk->Text);
	ini->WriteString("AUTO","zn_Ti",EznCti->Text);
	ini->WriteString("AUTO","zn_Td",EznCtd->Text);

	// model step graph
	ini->WriteInteger("STEP","max_time",MSForm->UDlen->Position);
}






















