//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - PWM generating module
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#include <string.h>

#pragma hdrstop

#include "pwm.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

TPWM pwm;

//---------------------------------------------------------------------------
// PWM - init
//---------------------------------------------------------------------------
void pwm_init(void)
{
	pwm.pos=0;
	pwm.duty=PWM_WAVE_SIZE/2;
	pwm.rst=0;
	pwm.inv=0;
}

//---------------------------------------------------------------------------
// PWM - set duty cycle
//---------------------------------------------------------------------------
void pwm_set_duty(double duty)
{
	// store duty
	pwm.duty=PWM_WAVE_SIZE*duty;
	if(pwm.duty>PWM_WAVE_SIZE-1)
		pwm.duty=PWM_WAVE_SIZE-1;
	else if(pwm.duty<1)
		pwm.duty=1;

	// calculates PWM levels
	if(pwm.duty>=0.5)
	{
		pwm.leva=0x00;
		pwm.levb=0x80 + 0x80*(1-pwm.duty)/pwm.duty;
	}
	else
	{
		pwm.leva=0x80 - 0x80*pwm.duty/(1-pwm.duty);
		pwm.levb=0xFF;
	}
	if(pwm.leva<0x00)
		pwm.leva=0x00;
	else if(pwm.leva>0x70)
		pwm.leva=0x70;
	if(pwm.levb>0xFF)
		pwm.levb=0xFF;
	else if(pwm.levb<0x90)
		pwm.levb=0x90;
}

//---------------------------------------------------------------------------
// PWM - get last duty cycle
//---------------------------------------------------------------------------
void pwm_get_duty(double *duty)
{
	if(!duty)
		return;
	*duty=(double)pwm.duty/PWM_WAVE_SIZE;
}

//---------------------------------------------------------------------------
// PWM - set generator frequency
//---------------------------------------------------------------------------
void pwm_set_freq(int freq)
{
	pwm.freq=freq;
}

//---------------------------------------------------------------------------
// PWM - set wave invert mode
//---------------------------------------------------------------------------
void pwm_set_invert(int inv)
{
	pwm.inv=inv;
}

//---------------------------------------------------------------------------
// PWM - generate PWM waveform at current frequency with current duty cycle
//---------------------------------------------------------------------------
int pwm_gen(char *data,int count,int srate)
{
	if(!data)
		return(1);

	// get resampling coefficient
	int step=(PWM_WAVE_SIZE*pwm.freq)/srate;

	// generate wave
	for(int i=0;i<count;i++)
	{
		uchar out;

		if(pwm.pos<pwm.duty)
			out=pwm.levb;
		else
			out=pwm.leva;

		if(pwm.rst==1)
		{
			out=pwm.leva;
			pwm.rst=2;
		}
		else if(pwm.rst==2)
		{
			out=pwm.levb;
			pwm.rst=0;
		}

		pwm.pos+=step;
		if(pwm.pos>=PWM_WAVE_SIZE)
		{
			pwm.pos-=PWM_WAVE_SIZE;
			pwm.rst=1;
			//out=0x00;
		}

    // invert waveform?
		if(pwm.inv)
			out=0xFF-out;

		*data++=(char)out;
		*data++=(char)out;
	}

	return(0);
}
