//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - discrete beta-I-PD regulator module
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#include <math.h>

#pragma hdrstop

#include "regulator.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

TReg reg;


//---------------------------------------------------------------------------
// Regulator - initialize
//---------------------------------------------------------------------------
void reg_init(int T)
{
	reg.sum=0.0;
	reg.dif=0.0;
	reg.x=0.0;
	reg.w=0;
	reg.T=(double)T/1000;

	reg.man=0;
	reg.sym=0;

	// default constants
	reg.k=1.0;
	reg.Ti=1.0;
	reg.Td=1.0;
	reg.N=1.0;
	reg.beta=1.0;
	reg.Tt=1.0;
}

//---------------------------------------------------------------------------
// Regulator - set desired value
//---------------------------------------------------------------------------
void reg_set_w(double w)
{
	reg.w=w;
}

//---------------------------------------------------------------------------
// Regulator - get desired value
//---------------------------------------------------------------------------
void reg_get_w(double *w)
{
	if(!w)
		return;
	*w=reg.w;
}

//---------------------------------------------------------------------------
// Regulator - get regulator command
//---------------------------------------------------------------------------
void reg_get_x(double *x)
{
	if(!x)
		return;
	*x=reg.x;
}

//---------------------------------------------------------------------------
// Regulator - get limitation flag
//---------------------------------------------------------------------------
int reg_get_x_limit(void)
{
	return(reg.lim);
}

//---------------------------------------------------------------------------
// Regulator - set manual mode
//---------------------------------------------------------------------------
void reg_set_man(int man,double duty)
{
	reg.man=man;
	if(man)
		reg.x=duty;
}

//---------------------------------------------------------------------------
// Regulator - get manual mode
//---------------------------------------------------------------------------
int reg_get_man(void)
{
	return(reg.man);
}

//---------------------------------------------------------------------------
// Regulator - get sampling period in seconds
//---------------------------------------------------------------------------
void reg_get_T(double *Ts)
{
	if(!Ts)
		return;
	*Ts=reg.T;
}

//---------------------------------------------------------------------------
// Regulator - set main constants
//---------------------------------------------------------------------------
void reg_set_PID(double k,double Ti,double Td)
{
	reg.k=k;
	reg.Ti=Ti;
	reg.Td=Td;
}

//---------------------------------------------------------------------------
// Regulator - set all constants
//---------------------------------------------------------------------------
void reg_set_PID_ext(double k,double Ti,double Td,double N,double beta,double Tt)
{
	reg.k=k;
	reg.Ti=Ti;
	reg.Td=Td;
	reg.N=N;
	reg.beta=beta;
	reg.Tt=Tt;
}

//---------------------------------------------------------------------------
// Regulator - set output limit mode (symetrical/asymetrical)
//---------------------------------------------------------------------------
void reg_set_limit(int sym)
{
	reg.sym=sym;
}

//---------------------------------------------------------------------------
// Regulator - get output limit mode (symetrical/asymetrical)
//---------------------------------------------------------------------------
int reg_get_limit(void)
{
	return(reg.sym);
}

//---------------------------------------------------------------------------
// Regulator - calculate one regulator step
//
// Features:
//  - beta-S-PD architecture
//  - output limitation
//  - dynamic anti-windup
//  - Impulse-Area-Invariant derivative component
//
// Constants/variables:
//  y    - system output (sensor)
//  x    - command to system
//  w    - desired value
//  T    - sampling period [s]
//  k    - gain (P compononent)
//  Ti   - integrator constant (S component)
//  Tt   - integrator dynamic antiwindup constant
//  Td   - derivator constant (D component)
//  N    - derivator filter constant
//  beta - 'w' gain for P and D component input (1: ordinaty PID, 0: I-PD)
//---------------------------------------------------------------------------
void reg_update(double y,double *x)
{
	double duty=reg.x;
	if(!reg.man)
	{
		// --- auto mode ---
		// regulation error
		double e = (double)(reg.w - y);
		double e_beta = reg.beta*(double)reg.w - (double)y;

		// new derivative component value IAI aproximation
		double newdiff = reg.dif*exp(-reg.T*reg.N/reg.Td) + e_beta*reg.k*reg.Td/reg.T*(1-exp(-reg.T*reg.N/reg.Td));

		// calculate regulator components
		double rP = reg.k*e_beta;
		double rS = reg.sum;
		double rD = newdiff - reg.dif;

		// get regulator output
		duty = rP + rS + rD;
		double duty_unlim=duty;
		if(duty>1)
			duty=1;
		else if(reg.sym && duty<-1)
			duty=-1;
		else if(duty<0)
			duty=0;

    // set limitation flag
		reg.lim=(duty!=duty_unlim);

		// update sumator output with dynamic anti-windup
		reg.sum += e*reg.k*reg.T/reg.Ti + (duty-duty_unlim)*reg.k*reg.T/reg.Tt;

		// hard limit sumator just in case of insane Tt constant value
		if(reg.sum>10)
			reg.sum=10;
		else if(reg.sum<-10)
			reg.sum=-10;



		// update derivator
		reg.dif = newdiff;
	}
	else
	{
		// --- manual mode ---

		// no limit
		reg.lim=0;

		// regulation error
		double e_beta = reg.beta*(double)reg.w - (double)y;

		// follow manually setup 'x'
		reg.sum=(reg.x-reg.k*e_beta);
		reg.dif=0;
	}

	// return command if requested
	reg.x=duty;
	if(x)
		*x=duty;
}


