//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - Sound card input/output module
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------


#include <windows.h>
#include <mmsystem.h>
#include <stdlib.h>

#pragma hdrstop

#include "sb_inout.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

double t_del=0.0;
int srate=0;
int per=0;
int end=0;
HWAVEIN sbin=NULL;
WAVEHDR *binhdr=NULL;
char *bufin=NULL;
int bufinlen=0;
int binused=0;
void (*fuin)(UINT,char*,int,int)=NULL;
void (*fuout)(UINT,char*,int,int)=NULL;

HWAVEOUT sbout=NULL;
WAVEHDR *bouthdr=NULL;
char *bout=NULL;
int boutlen=0;
int boutused=0;

//---------------------------------------------------------------------------
// Close SB
//---------------------------------------------------------------------------
void sb_close(void)
{
	// ====== INPUT ======
	if(sbin)
	{
		// close input if opened

		// end flag
		end=1;

		// release buffers
		//waveInReset(sbin);

		// wait while not finished
		while(binused)
			Sleep(10);

		// close input device
		if(waveInClose(sbin)!=MMSYSERR_NOERROR)
			return;
		sbin=NULL;
	}

	// loose input data buffers
	if(bufin)
		free((void*)bufin);
	bufin=NULL;

	// loose input buffer headers
	if(binhdr)
		free((void*)binhdr);
	binhdr=NULL;



	// ====== OUTPUT ======
	if(sbout)
	{
		// close output if opened

		// end flag if not set yet
		end=1;

		// wait until all buffers returned
		while(boutused)
			Sleep(10);

		// close output device
		if(waveOutClose(sbout)!=MMSYSERR_NOERROR)
			return;
		sbout=NULL;
	}


	// loose output data buffers
	if(bout)
		free((void*)bout);
	bout=NULL;

	// loose output buffer headers
	if(bouthdr)
		free((void*)bouthdr);
	bouthdr=NULL;
}

//---------------------------------------------------------------------------
// SB - get output devices count
//---------------------------------------------------------------------------
void sb_inout_dev_count(int *in,int *out)
{
	if(in)
		*in=waveInGetNumDevs();
	if(out)
		*out=waveOutGetNumDevs();
}

//---------------------------------------------------------------------------
// SB - get output device parameters
//---------------------------------------------------------------------------
int sb_get_output_dev(int out,int id,char *name)
{
	if(!name)
		return(1);

	// get device caps
	WAVEOUTCAPS cout;
	WAVEINCAPS cin;
	int err;
	if(out)
		err=waveOutGetDevCaps((UINT_PTR)id,&cout,sizeof(WAVEOUTCAPS));
	else
		err=waveInGetDevCaps((UINT_PTR)id,&cin,sizeof(WAVEINCAPS));
	if(err!=MMSYSERR_NOERROR)
		return(2);

	// return device name
	if(out)
		strcpy(name,cout.szPname);
	else
		strcpy(name,cin.szPname);

	return(0);
}

//---------------------------------------------------------------------------
// Get SB propagation delay
//---------------------------------------------------------------------------
double sb_get_delay(void)
{
	return(t_del);
}

//---------------------------------------------------------------------------
// Init SB
//---------------------------------------------------------------------------
int sb_init(int inid,int outid,int freq,int sper,int outdel,void (*func_in)(UINT,char*,int,int),void (*func_out)(UINT,char*,int,int))
{
	if(freq<8000 || freq>192000)
		return(1);

	if(sper<50 || sper>10000)
		return(1);

	// init variables
	end=0;
	binused=0;
	boutused=0;

	// total input-output propagation delay [periods]
	t_del=((double)outdel + (SBOUT_BUFN)*((double)sper)/2);

	// register input done user callback function
	fuin=func_in;

	// register output done user callback function
	fuout=func_out;

	// store sampling freq
	srate=freq;


	// setup sb format
	WAVEFORMATEX wave;
	wave.wFormatTag=WAVE_FORMAT_PCM;
	wave.nChannels=2;
	wave.wBitsPerSample=8;
	wave.nSamplesPerSec=freq;
	wave.nBlockAlign=wave.nChannels*wave.wBitsPerSample/8;
	wave.nAvgBytesPerSec=wave.nBlockAlign*wave.nSamplesPerSec;
	wave.cbSize=sizeof(WAVEFORMATEX);



	// =========================
	// ====== INPUT SETUP ======
	// =========================

	// calculate input buffer length (to be multiple of 2 samples)
  per=sper;
	bufinlen=((wave.nAvgBytesPerSec*per)/4000)*4;

	// allocate input data buffers
	bufin=(char*)malloc(bufinlen*SBIN_BUFN);
	if(!bufin)
	{
		// allocation failed
		return(2);
	}

	// allocate input buffer headers
	binhdr=(WAVEHDR*)malloc(sizeof(WAVEHDR)*SBIN_BUFN);
	if(!binhdr)
	{
		// allocation failed
		sb_close();
		return(2);
	}

	// open input device
	MMRESULT err=waveInOpen(&sbin,(UINT_PTR)inid,&wave,(DWORD_PTR)waveInProc,0,CALLBACK_FUNCTION);
	if(err!=MMSYSERR_NOERROR)
	{
		// failed
		sb_close();
		return(3);
	}

	// init input buffers
	for(int i=0;i<SBIN_BUFN;i++)
	{
		binhdr[i].lpData=&bufin[i*bufinlen];
		binhdr[i].dwBufferLength=bufinlen;
		binhdr[i].dwBytesRecorded=0;
		binhdr[i].dwUser=0;
		binhdr[i].dwFlags=0;
		binhdr[i].dwLoops=0;
		binhdr[i].lpNext=NULL;
	}

	// prepare them and add them to the device
	for(int i=0;i<SBIN_BUFN;i++)
	{
		waveInPrepareHeader(sbin,&binhdr[i],sizeof(WAVEHDR));
		waveInAddBuffer(sbin,&binhdr[i],sizeof(WAVEHDR));
		binused++;
	}



	// ==========================
	// ====== OUTPUT SETUP ======
	// ==========================

	// output buffer size (half of input buffer length)
	boutlen=bufinlen/2;
	// initial size to ensure phase input/output shift
	int binitlen=boutlen + ((wave.nAvgBytesPerSec*outdel)/2000)*2;
	// total buffer length
	int len=(binitlen>boutlen*SBOUT_BUFN)?binitlen:(boutlen*SBOUT_BUFN);

	// allocate output data buffers
	bout=(char*)malloc(len);
	if(!bout)
	{
		// allocation failed
		sb_close();
		return(2);
	}
	// clear buffers
	memset((void*)bout,0x80,boutlen*SBOUT_BUFN);

	// allocate output buffer headers
	bouthdr=(WAVEHDR*)malloc(sizeof(WAVEHDR)*SBOUT_BUFN);
	if(!bouthdr)
	{
		// allocation failed
		sb_close();
		return(2);
	}


	// create output thread
  ULONG tid;
	HANDLE thout=CreateThread(NULL,1024,waveOutProc,NULL,0,&tid);
	if(!thout)
	{
		// output thread creation failed
		sb_close();
		return(4);
	}

	// increase priority
	SetThreadPriority(thout,THREAD_PRIORITY_ABOVE_NORMAL);

	// close handle to the thread (it still exists out there :-)
	CloseHandle(thout);



	// open output device
	err=waveOutOpen(&sbout,(UINT_PTR)outid,&wave,(DWORD_PTR)tid,0,CALLBACK_THREAD);
	if(err!=MMSYSERR_NOERROR)
	{
		// failed
		sb_close();
		return(4);
	}

	// init output buffers
	for(int i=0;i<SBOUT_BUFN;i++)
	{
		bouthdr[i].lpData=&bout[i*boutlen];
		bouthdr[i].dwBufferLength=boutlen;
		bouthdr[i].dwBytesRecorded=0;
		bouthdr[i].dwUser=0;
		bouthdr[i].dwFlags=0;
		bouthdr[i].dwLoops=0;
		bouthdr[i].lpNext=NULL;
	}
	// first buffer to half size (T/2 phase shift relative to input device)
	bouthdr[0].dwBufferLength=binitlen;

	// pause output
	waveOutPause(sbout);

	// prepare them and add them to the device
	for(int i=0;i<SBOUT_BUFN;i++)
	{
		waveOutPrepareHeader(sbout,&bouthdr[i],sizeof(WAVEHDR));
		waveOutWrite(sbout,&bouthdr[i],sizeof(WAVEHDR));
		boutused++;
	}

	// start both in/out
	waveInStart(sbin);
	waveOutRestart(sbout);

	return(0);
}

//---------------------------------------------------------------------------
// SB input - get input buffer length
//---------------------------------------------------------------------------
int sb_get_input_len(int *len)
{
	if(!len)
		return(1);

	*len=bufinlen;
	return(0);
}






//---------------------------------------------------------------------------
// SB input callback
//---------------------------------------------------------------------------
void CALLBACK waveInProc(HWAVEIN hwi,UINT uMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2)
{
	WAVEHDR *pWaveHdr;

	switch(uMsg)
	{
		case WIM_OPEN:
			// ====== SB INPUT OPEN ======

			// call user registered function
			if(fuin)
				fuin(uMsg,NULL,0,srate);
			break;
		case WIM_CLOSE:
			// ====== SB INPUT CLOSE ======

			// call user registered function
			if(fuin)
				fuin(uMsg,NULL,0,srate);
			break;
		case WIM_DATA:
			// ====== SB INPUT DATA READY ======

			// returned buffer
			pWaveHdr=(WAVEHDR*)(dwParam1);

			// unprepare buffer header
			waveInUnprepareHeader(sbin,pWaveHdr,sizeof(WAVEHDR));
			binused--;

			if(!end)
			{
				// call user registered function
				if(fuin)
					fuin(uMsg,pWaveHdr->lpData,pWaveHdr->dwBytesRecorded/2,srate);

				// send in new buffer
				pWaveHdr->dwUser=0;
				pWaveHdr->dwFlags=0;
				pWaveHdr->dwLoops=0;
				pWaveHdr->lpNext=NULL;
				pWaveHdr->reserved=0;
				waveInPrepareHeader(sbin,pWaveHdr,sizeof(WAVEHDR));
				waveInAddBuffer(sbin,pWaveHdr,sizeof(WAVEHDR));
				binused++;
			}
			break;
	}
}

//---------------------------------------------------------------------------
// SB output callback thread
//---------------------------------------------------------------------------
DWORD WINAPI waveOutProc(LPVOID arg)
{
	MSG msg;

	// wait for some message
	while(GetMessage(&msg,NULL,0,0)>0)
	{
		switch(msg.message)
		{
			case WOM_OPEN:
				// ====== SB OUTPUT OPEN ======

				// call user registered function
				if(fuout)
					fuout(msg.message,NULL,0,srate);
				break;
			case WOM_CLOSE:
				// ====== SB OUTPUT CLOSE ======

				// call user registered function
				if(fuout)
					fuout(msg.message,NULL,0,srate);
				break;
			case WOM_DONE:
				// ====== SB OUTPUT DATA READY ======

				// finished buffer
				WAVEHDR *pWaveHdr=(WAVEHDR*)(msg.lParam);

				// unprepare buffer header
				waveOutUnprepareHeader(sbout,pWaveHdr,sizeof(WAVEHDR));
				boutused--;

				if(!end)
				{
					// call user registered function
					if(fuout)
						fuout(msg.message,pWaveHdr->lpData,pWaveHdr->dwBufferLength/2,srate);
          
					// send in new buffer
					pWaveHdr->dwUser=0;
					pWaveHdr->dwFlags=0;
					pWaveHdr->dwLoops=0;
					pWaveHdr->lpNext=NULL;
					pWaveHdr->reserved=0;
					// buffer length to full size
					pWaveHdr->dwBufferLength=boutlen;

					waveOutPrepareHeader(sbout,pWaveHdr,sizeof(WAVEHDR));
					waveOutWrite(sbout,pWaveHdr,sizeof(WAVEHDR));
					boutused++;
				}
				break;

		}
	}

	return(0);
}
