//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - process identification module
//                                        - auto-tuning
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#pragma hdrstop

#include "sys_model.h"
#include "regulator.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)


TSysId sys;

//---------------------------------------------------------------------------
// System identificator - initialization
//---------------------------------------------------------------------------
void sys_idn_init(void)
{
	for(int i=0;i<SYS_N;i++)
		sys.X[i]=NULL;
	sys.T=NULL;

	sys.pmax=0;
	sys.n=0;
	sys.ppos=0;
	sys.filter=0;
	sys.itmax=0;
	sys.err=0.0;

	sys.enab=1;

	for(int i=0;i<SYS_N;i++)
		sys.th[i]=0.0;

	sys.ferr=0.0;
	sys.fcyc=0;

	sys.zn_adapt=0;
}

//---------------------------------------------------------------------------
// System identificator - close data buffers
//---------------------------------------------------------------------------
void sys_idn_close(void)
{
	for(int i=0;i<SYS_N;i++)
		free((void*)sys.X[i]);
	free((void*)sys.T);
	for(int i=0;i<SYS_N;i++)
		sys.X[i]=NULL;
	sys.T=NULL;

	sys.ppos=0;
	sys.pmax=0;
	sys.n=0;
}

//---------------------------------------------------------------------------
// System identificator - set initial patterns count and sampling rate
//---------------------------------------------------------------------------
int sys_idn_config(int pat,double Ts)
{
	int i;

	if(pat!=sys.pmax)
	{
		// loose old buffers
		sys_idn_close();

		// allocate pattern matrix 'X'
		for(i=0;i<SYS_N;i++)
		{
			sys.X[i]=(double*)malloc(pat*sizeof(double));
			if(!sys.X[i])
			{
				sys_idn_close();
				return(1);
			}
		}

		// output samples vector 'T'
		sys.T=(double*)malloc(pat*sizeof(double));
		if(!sys.T)
		{
			sys_idn_close();
			return(1);
		}

		// initial weights 'w'
		sys_idn_reset(1);

		// store max patterns count
		sys.pmax=pat;

		// no pattern/input buffered yet
		sys.n=0;

		// round buffer X and T position
		sys.ppos=0;
	}

	// store sampling T
	if(Ts>1E-9)
		sys.Ts=Ts;

	return(0);
}

//---------------------------------------------------------------------------
// System identificator - reset model
//---------------------------------------------------------------------------
void sys_idn_reset(int filter)
{
	int i;
	
	// initial weights 'w'
	for(i=0;i<SYS_N;i++)
		sys.w[i]=0.0;
	sys.w[SYS_D]=1.0;

  // clear filtered model to?
	if(filter)
		memcpy((void*)sys.th,(void*)sys.w,SYS_N*sizeof(double));

	// clear patterns
	sys.n=0;
	sys.ppos=0;
}

//---------------------------------------------------------------------------
// System identificator - limit learning algorithm cycles count per call
//                      - set final learning algorithm error
//---------------------------------------------------------------------------
void sys_idn_limit(int cycles,double error)
{
	sys.itmax=cycles;
	sys.err=error;
}

//---------------------------------------------------------------------------
// System identificator - get learning algorithm limits
//---------------------------------------------------------------------------
void sys_idn_limit(int *cycles,double *error)
{
	if(cycles)
		*cycles=sys.itmax;

	if(error)
		*error=sys.err;
}

//---------------------------------------------------------------------------
// System identificator - set output filter periods count
//---------------------------------------------------------------------------
void sys_idn_filter(int len)
{
	sys.filter=len;
}

//---------------------------------------------------------------------------
// System identificator - enable/disable identification
//---------------------------------------------------------------------------
void sys_idn_state(int enabled)
{
	sys.enab=enabled;
}

//---------------------------------------------------------------------------
// System identificator - enable/disable adaptive mode (just internal variable)
//---------------------------------------------------------------------------
int sys_idn_adaptive(void)
{
	return(sys_idn_adaptive(-1));
}
int sys_idn_adaptive(int enabled)
{
	// get last state
	int ret=sys.zn_adapt;

	// set new if requested
	if(enabled>=0)
		sys.zn_adapt=enabled;

	// return
	return(ret);
}

//---------------------------------------------------------------------------
// System identificator - get identified ARX model, output prediction,
// closed loop critical parameters, algorithm error and used cycles count
//---------------------------------------------------------------------------
int sys_idn_get_model(double *ypred,double *theta,double *kp,double *err,int *cycles)
{
	// model prediction
	if(ypred)
		*ypred=sys.y_pred;

	// model parameters (b0, b1, ... , bn, a0, a1, ..., an)
	if(theta)
		memcpy((void*)theta,(void*)sys.th,SYS_N*sizeof(double));

	// closed loop critical parameters (Kkrit, Tkrit)
	if(kp)
		memcpy((void*)kp,(void*)sys.kp,2*sizeof(double));

	// model error
	if(err)
		*err=sys.ferr;

	// SD cycles
	if(cycles)
		*cycles=sys.fcyc;

	return(sys.enab);
}

//---------------------------------------------------------------------------
// System identificator - get gain of the identified ARX model
//---------------------------------------------------------------------------
double sys_idn_get_model_gain(void)
{
	// model gain: sum(bi,i=1..n)/sum(ai,i=1..n)
	double nom=0.0;
	double den=1.0;
	for(int i=0;i<SYS_N/2;i++)
	{
		nom+=sys.th[i];
		den+=sys.th[i+SYS_N/2];
	}
	if(den==0.0)
		return(0.0);
	else
		return(nom/den);
}

//---------------------------------------------------------------------------
// System identificator - main identification algorithm
//
// Neuro-ARX system model with Steepest-Descent learning algoritm.
//
// Description
// -----------
// nx: system input sample x[k]
// ny: system output sample y[k]
//
// system model:
// y[k] = b1*x[k-1] + b2*x[k-2] + ... + bn*x[k-n] - a1*y[k-1] - a2*y[k-2] - an*y[k-n],
// where neuron weights are w[] = (b1,b2,...,bn,a1,a2,...,an)^T
//
// T vector: stores undelayed output values (y[k],y[k-1],y[k-2],...)^T
// X matrix: stores patterns to learn and at algorithm start looks like this:
//   (x[k-1],x[k-2],...,-y[k-1],-y[k-2],...)
//   (x[k-2],x[k-3],...,-y[k-2],-y[k-3],...)
//   (x[k-3],x[k-4],...,-y[k-3],-y[k-4],...)
//   (.....................................)
// Note: I'm using round buffers for X and T instead of shifting its content. 
//
// Main SD algorithm:
// ------------------
// Jacobian matrix:
//   J = -X
// Hessian matrix:
//   H = J'*J
//
// iter:
//   current prediction error:
//     e = T - X*w
//
//   gradient vector:
//     g = J'*e
//
//   weights correction coefficient:
//     eta = (g'*H*g)^-1*g'*g
//
//   update weights:
//     w = w - eta*g
//
//   get model error:
//     err = e'*e/n
//
//   if err>err_stop and not max_cycles then goto iter
//---------------------------------------------------------------------------
int sys_idn_update(double nx,double ny,double *ypred,double *theta,double *kp)
{
	int i,j,k;

	// leave if no buffers available
	if(!sys.pmax || !sys.enab)
		return(1);


	// === add new sample into output samples buffer T ===
	// add new output sample
	sys.T[sys.ppos]=ny;


	// === STEEPEST-DESCENT ALGORITHM ===
	if(sys.n>=SYS_N*2)
	{
		// jacobian matrix: J = -X

		// calculate hessian matrix: J'*J
		double H[SYS_N][SYS_N];
		for(i=SYS_D;i<SYS_N;i++)
		{
			double *row=sys.X[i];
			for(j=SYS_D;j<SYS_N;j++)
			{
				double *col=sys.X[j];
				double sum=0.0;
				for(k=0;k<sys.n;k++)
					sum+=row[k]*col[k];
				H[i][j]=sum;
			}
		}

		// --- for each iteration ---
		double err=0.0;
		for(i=0;i<sys.itmax;i++)
		{
			// error vector: E = T - X*w
			// gradient: G = J'*E
			double G[SYS_N];
			for(k=SYS_D;k<SYS_N;k++)
				G[k]=0.0;
			// iteration error: err = E'*E/n

			// calculate: E, G, err
			err=0.0;
			for(j=0;j<sys.n;j++)
			{
				// E(j)
				double pre=0.0;
				for(k=SYS_D;k<SYS_N;k++)
					pre+=sys.X[k][j]*sys.w[k];

				// E(j) = T(j) - X(j,:)*w
				double E=sys.T[j]-pre;

				// G += J(j,:)'*E(j) = -X(j,:)'*E(j)
				for(k=SYS_D;k<SYS_N;k++)
					G[k]-=sys.X[k][j]*E;

				// err += E(j)^2
				err+=E*E;
			}
			err/=(double)sys.n;

			// calculate weight correction coeff.: eta = G'*G/(G'*JJ*G)
			double etad=0.0;
			for(j=SYS_D;j<SYS_N;j++)
			{
				double gjj=0.0;
				for(k=SYS_D;k<SYS_N;k++)
					gjj+=G[k]*H[k][j];
				etad+=gjj*G[j];
			}
			// leave if invalid op.
			if(etad==0.0)
				break;
			double eta=0.0;
			for(k=SYS_D;k<SYS_N;k++)
				eta+=G[k]*G[k];
			eta/=etad;

			// update model weights: w = w - eta*G
			for(k=SYS_D;k<SYS_N;k++)
				sys.w[k]-=eta*G[k];

			// done if error low enough
			if(err<sys.err)
				break;
		}

		// store last error
		sys.ferr=err;

		// and cycles used
		sys.fcyc=(i>=sys.itmax)?sys.itmax:(i+1);
	}

	// --- filter model constants ---
	for(k=SYS_D;k<SYS_N;k++)
		sys.th[k]=(sys.th[k]*(double)(sys.filter-1) + sys.w[k])/(double)sys.filter;


	// === add new pattern into pattern buffer ===
	// --- update position in X and T round buffers ---
	int lpos=sys.ppos;
	sys.ppos++;
	if(sys.ppos>=sys.pmax)
		sys.ppos=0;

	// copy last pattern to new position
	for(i=0;i<SYS_N;i++)
		sys.X[i][sys.ppos]=sys.X[i][lpos];

	// shift current pattern data by one sample in time
	for(i=SYS_N-1;i>0;i--)
		sys.X[i][sys.ppos]=sys.X[i-1][sys.ppos];
	// add new input sample
	sys.X[0][sys.ppos]=nx;
	// and new output sample
	sys.X[SYS_N/2][sys.ppos]=-ny;


	// --- model prediction ---
	double y_pred=0.0;
	for(k=0;k<SYS_N;k++)
		y_pred+=sys.th[k]*sys.X[k][lpos];

	// limit model output
	if(y_pred>1000)
		y_pred=1000;
	else if(y_pred<-1000)
		y_pred=-1000;

	// return if requested
	sys.y_pred=y_pred;
	if(ypred)
		*ypred=y_pred;

	// --- increase pattern count if maximum not reached yet ---
	sys.n++;
	if(sys.n>sys.pmax)
		sys.n=sys.pmax;

	// --- return ARX parameters if requested ---
	if(theta)
		memcpy((void*)theta,(void*)sys.th,SYS_N*sizeof(double));

	// --- calculate poles and zeros for the model ---
	sys_idn_eval_PZ(sys.th,sys.mpole,&sys.mpolen,sys.mzero,&sys.mzeron);

	// --- calculate critical paramters for system ---
	sys_idn_get_critical(sys.th,sys.kp,sys.cpole);

	// return them if requested
	if(kp)
		memcpy((void*)kp,(void*)sys.kp,sizeof(double)*2);


	return(0);
}


//---------------------------------------------------------------------------
// System identification - simulate model for input x for n periods 
//---------------------------------------------------------------------------
int sys_idn_step(double *th,double x,double *y,int n)
{
	return(sys_idn_lsim(th,&x,y,n,LSIM_STEP));
}
int sys_idn_pulse(double *th,double x,double *y,int n)
{
	return(sys_idn_lsim(th,&x,y,n,LSIM_PULSE));
}
int sys_idn_lsim(double *th,double *x,double *y,int n)
{
  return(sys_idn_lsim(th,x,y,n,LSIM_GEN));
}
int sys_idn_lsim(double *th,double *x,double *y,int n,unsigned int mode)
{
	int i,j;

	if(!th || !x || !y || mode>LSIM_PULSE)
		return(1);

	// clear memory
	double mem[SYS_N];
	for(i=0;i<SYS_N;i++)
		mem[i]=0.0;

	// simulate system for n samples
	for(i=0;i<n;i++)
	{
		// calculate output
		y[i]=0.0;
		for(j=0;j<SYS_N;j++)
			y[i]+=th[j]*mem[j];

		// shift register
		for(j=SYS_N-1;j>0;j--)
			mem[j]=mem[j-1];
		if(mode==LSIM_GEN)
			mem[0]=x[i];
		else if(mode==LSIM_STEP)
			mem[0]=x[0];
		else
      mem[0]=(i==0)?x[0]:0;
		mem[SYS_N/2]=-y[i];
	}

	return(0);
}


//---------------------------------------------------------------------------
// System identification - print system equation
//---------------------------------------------------------------------------
int sys_idn_print_system(char *str,int mode)
{
	if(!str)
		return(1);

	if(mode==1)
		sprintf(str,"Fz = tf([0 %.5f %.5f %.5f],[1 %.5f %.5f %.5f],%.3f)",sys.th[0],sys.th[1],sys.th[2],sys.th[3],sys.th[4],sys.th[5],sys.Ts);
	else
		sprintf(str,"z = tf(\'z\',%.3f)\nFz = (%+.5f*z^-1%+.5f*z^-2%+.5f*z^-3)/(1%+.5f*z^-1%+.5f*z^-2%+.5f*z^-3)",sys.Ts,sys.th[0],sys.th[1],sys.th[2],sys.th[3],sys.th[4],sys.th[5]);
	return(0);
}

//---------------------------------------------------------------------------
// System identification - calculates system critical parameters for closed loop
// for 3rd order system.
//
// This piece of code is mostly recycled from my old college project,
// I hope it's ok. :-)
//---------------------------------------------------------------------------
// Critical parameters are calculated from closed loop characteristic polynomial:
// D(z) = A + B*Kk = 1 + z^-1*(a1 + b1*Kk) + z^-2*(a2 + b2*Kk) + z^-3*(a3 + b3*Kk)
//
// In one case D contains complex pair of poles at unit circle and one stable inside:
// P1(z) = (z - al +/- beta)*(1 + e1*z^-1), al^2 + beta^2 = 1, |e1|<1
// P1(z) = (1 - 2*al*z^-1 + z^-2)*(1 + e1*z^-1)
//
// From D(z) = P1(z):
// r0 = a3*(a3 - a1) + a2 - 1
// r1 = b3*(2*a3 - a1) + b2 - a3*b1
// r2 = b3*(b3 - b1)
// e1? = a3 + b3*Kk
// Kk1,2 = (-r1 +/- sqrt(r1^2 - 4*r0*r2))/(2*r2)
// al = (a3 - a1 + Kk*(b3 - b1))/2
// Tk = 2*pi*Ts/arccos(al)
//
// One of these solutions is valid when |e1?|<1 and Kk?>0. If not, the D might
// contain one pole at -1:
// P2(z) = (1 + z^-1)*(1 + f1*z^-1 + f2*z^-2)
//
// From D(z) = P1(z):
// f1 = (a1*(b3 - b2) - b3 + b2 + (a2 - a3)*b1)/(b3 - b2 + b1)
// f2 = (a2*b3 - a1*b3 + b3 - a3*b2 + a3*b1)/(b3 - b2 + b1)
// Kk = (1 - a1 + a2 - a3)/(b1 - b2 + b3)
// Tk = 2*Ts
//---------------------------------------------------------------------------
int sys_idn_get_critical(double *theta,double *kp,TPoZe *pole)
{
	if(!theta)
		return(1);

	// ARX model parameters for order 3 system
	double b1=theta[0];
	double b2=theta[1];
	double b3=theta[2];
	double a1=theta[3];
	double a2=theta[4];
	double a3=theta[5];

	// first solution coeffs.
	double r0=a3*(a3-a1)+a2-1;
	double r1=b3*(2*a3-a1)+b2-a3*b1;
	double r2=b3*(b3-b1);
	double d=r1*r1-4*r0*r2;
	if(d<0)
		d=0;

	// fix /0 cases
	if(r2==0.0)
		r2+=1E-9;
	if((b1-b2+b3)==0.0)
		b1+=1E-9;

	// critical gain solutions
	double Kp1=(-r1+sqrt(d))/2/r2;
	double Kp2=(-r1-sqrt(d))/2/r2;
	double Kp3=(1-a1+a2-a3)/(b1-b2+b3);

	// conditions
	double e11=a3+b3*Kp1;
	double e12=a3+b3*Kp2;

	// choose solution and calculate oscillation period
	double e1;
	double Tk=0.0;
	double Kpk;
	if(Kp1>0 && fabs(e11)<1)
	{
		// P1(z), 1. solution
		Kpk=Kp1;
		e1=e11;
	}
	else if(Kp2>0 && fabs(e12)<1)
	{
		// P1(z), 2. solution
		Kpk=Kp2;
		e1=e12;
	}
	else
	{
		// P2(z)
		Kpk=Kp3;
		Tk=2*sys.Ts;

		// calc. poles
		if(pole)
		{
			double f1=(a1*(b3-b2)-b3+b2+(a2-a3)*b1)/(b3-b2+b1);
			double f2=(a2*b3-a1*b3+b3-a3*b2+a3*b1)/(b3-b2+b1);
			double Df=f1*f1-4*f2;
			if(Df>=0.0)
			{
				pole[0].re=(-f1+sqrt(Df))/2;
				pole[0].im=0.0;
				pole[1].re=(-f1-sqrt(Df))/2;
				pole[1].im=0.0;
			}
			else
			{
				pole[0].re=-f1/2;
				pole[0].im=sqrt(-Df)/2;
				pole[1].re=pole[0].re;
				pole[1].im=-pole[0].im;
			}
			pole[2].re=-1.0;
			pole[2].im=0.0;
		}
	}
	if(Tk==0.0)
	{
		// P1(z)
		double al=(a3-a1+Kpk*(b3-b1))/2;
		if(al>1.0)
			al=1.0;
		else if(al<-1.0)
			al=-1.0;

		double wk=1.0/sys.Ts*acos(al);
		if(wk==0)
			wk=1E-9;
		Tk=2*3.1415/wk;

		// calc. poles
		if(pole)
		{
			pole[0].re=al;
			pole[0].im=sqrt(1-al*al);
			pole[1].re=al;
			pole[1].im=-sqrt(1-al*al);
			pole[2].re=-e1;
			pole[2].im=0.0;
		}
	}

	// and here we go: critical gain and oscillation period (hopefully :-)
	Kpk=fabs(Kpk);
	Tk=fabs(Tk);

	// return critical parameters if requested
	if(kp)
	{
		kp[0]=Kpk;
		kp[1]=Tk;
	}

	return(0);
}


//---------------------------------------------------------------------------
// System identification - return poles for critical gain
//---------------------------------------------------------------------------
void sys_idn_get_poles(TPoZe *pole)
{
	if(!pole)
		return;

	memcpy((void*)pole,(void*)sys.cpole,3*2*sizeof(double));
}

//---------------------------------------------------------------------------
// System identification - calculate poles and zeros for third order system
//
// Cubic equation solution according to:
// http://www.1728.org/cubic2.htm
//---------------------------------------------------------------------------
void sys_idn_eval_PZ_solve(double *pl,TPoZe *rt,int *n)
{
	if(!pl || !rt || !n)
		return;

	if(pl[0]!=0)
	{
		// === CUBIC EQUATION ===
		// a*x^3 + b*x^2 + c*x^1 + d = 0

		// f = ((3c/a) - (b^2/a^2))/3
		double f = (3*pl[2]/pl[0] - pl[1]*pl[1]/(pl[0]*pl[0]))/3;

		// g = ((2b^3/a^3) - (9bc/a^2) + (27d/a))/27
		double g = (2*pl[1]*pl[1]*pl[1]/(pl[0]*pl[0]*pl[0]) - (9*pl[1]*pl[2]/(pl[0]*pl[0])) + (27*pl[3]/pl[0]))/27;

		// h = (g^2/4) + (f^3/27)
		double h = g*g/4 + f*f*f/27;

		double nr;
		if(h>0)
		{
			// --- one complex pair, one real ---

			// R = -g/2 + h^(1/2)
			double R = -g/2 + sqrt(h);

			// S = R^(1/3)
			nr=(R<0)?-1:1;
			double S = nr*pow(nr*R,1.0/3.0);

			// T = -g/2 - h^(1/2)
			double T = -g/2 - sqrt(h);

			// U = T^(1/3)
			nr=(T<0)?-1:1;
			double U = nr*pow(nr*T,1.0/3.0);

			// real root: (S+U) - b/(3*a)
			rt[0].re = (S+U) - pl[1]/(3*pl[0]);
			rt[0].im = 0;

			// complex pair: -(S+U)/2 - b/(3*a) +/- j*(S-U)*3^(1/2)/2
			rt[1].re = -(S+U)/2 - pl[1]/(3*pl[0]);
			rt[1].im = +(S-U)*1.732050808/2;
			rt[2].re = +rt[1].re;
			rt[2].im = -rt[1].im;

			// three roots
			*n=3;
		}
		else if(f+g+h==0)
		{
			// --- the same reals ---

			// r = -(d/a)^(1/3)
			double r = pl[3]/pl[0];
			nr = (r<0)?-1:1;
			r = -nr*pow(nr*r,1.0/3.0);

			// roots
			rt[0].re = r;
			rt[1].re = r;
			rt[2].re = r;
			rt[0].im = 0;
			rt[1].im = 0;
			rt[2].im = 0;

			// three roots
			*n=3;
		}
		else
		{
			// --- reals only ---

			// i = (g^2/4 - h)^(1/2)
			double i = sqrt(g*g/4 - h);

			// j = i^(1/3)
			nr=(i<0)?-1:1;
			double j = nr*pow(nr*i,1.0/3.0);

			// k = arccos(-g/(2*i))
			double k = acos(-g/(2*i));

			// L = -j
			double L = -j;

			// M = cos(k/3)
			double M = cos(k/3);

			// N = 1.732050808*sin(k/3)
			double N = 1.732050808*sin(k/3);

			// P = -b/3*a
			double P = -pl[1]/(3*pl[0]);

			// 1. root: 2*j*cos(k/3) - b/(3*a)
			rt[0].re = 2*j*cos(k/3) - pl[1]/(3*pl[0]);
			rt[0].im = 0;

			// 2. root: L*(M+N)+P
			rt[1].re = L*(M+N)+P;
			rt[1].im = 0;

			// 3. root: L*(M-N)+P
			rt[2].re = L*(M-N)+P;
			rt[2].im = 0;

			// three roots
			*n=3;
		}
	}
	else if(pl[1]!=0)
	{
		// === QUADRATIC EQUATION ===
		// a*x^2 + b*x^1 + c = 0

		// D = b^2 - 4*a*c
		double D = pl[2]*pl[2] - 4*pl[1]*pl[3];

		if(D<0)
		{
			// complex pair
			rt[0].re = -pl[2]/(2*pl[1]);
			rt[0].im = +sqrt(-D)/(2*pl[1]);
			rt[1].re = +rt[0].re;
			rt[1].im = -rt[0].im;
		}
		else
		{
			// reals
			rt[0].re = (-pl[2] + sqrt(D))/(2*pl[1]);
			rt[0].im = 0;
			rt[1].re = (-pl[2] + sqrt(D))/(2*pl[1]);
			rt[1].im = 0;
		}

		// two roots
		*n=2;
	}
	else if(pl[2]!=0)
	{
		// === LINEAR EQUATION ===
		// a*x + b = 0

		// root:
		rt[0].re = -pl[3]/pl[2];
		rt[0].im = 0;

		// single root
		*n=1;
	}
	else
	{
		// no roots
		*n=0;
	}
}
void sys_idn_eval_PZ(double *th,TPoZe *poles,int *pn,TPoZe *zeros,int *zn)
{
	double poly[4];

	// --- find poles ---
	if(poles && pn)
	{
		// make polynome for poles: z^3 + th[0]*z^2 + th[1]*z + th[2]
		poly[0]=1.0;
		memcpy((void*)&poly[1],(void*)&th[SYS_N/2],SYS_N/2*sizeof(double));

		// find roots
		sys_idn_eval_PZ_solve(poly,poles,pn);
	}

	// --- find zeros ---
	if(zeros && zn)
	{
		// make polynome for zeros: th[3]*z^2 + th[4]*z + th[5]
		poly[0]=0.0;
		memcpy((void*)&poly[1],(void*)&th[0],SYS_N/2*sizeof(double));

		// find roots
		sys_idn_eval_PZ_solve(poly,zeros,zn);
	}
}

//---------------------------------------------------------------------------
// System identification - return ARX model poles/zeros
//---------------------------------------------------------------------------
void sys_idn_get_PZ(TPoZe *poles,int *pn,TPoZe *zeros,int *zn)
{
	// return poles
	if(poles && pn)
	{
		memcpy((void*)poles,(void*)sys.mpole,SYS_N/2*sizeof(TPoZe));
		*pn=sys.mpolen;
	}

	// return zeros
	if(zeros && zn)
	{
		memcpy((void*)zeros,(void*)sys.mzero,SYS_N/2*sizeof(TPoZe));
		*zn=sys.mzeron;
	}
}


//---------------------------------------------------------------------------
// System identification - set coefficients for Zielgler-Nichols tune method
//
// ZN method then returns PID following tuning parameters:
// K  = zn_k*Kk,
// Ti = zn_Ti*Tk,
// Td = zn_Td*Tk,
// where Kk is closed loop critical gain and Tk oscillation period.
//
// Usuall settings: 0.6, 0.5, 0.125 (standard)
//                  0.3, 1.0, 0.125 (tu reduced ringing)
//---------------------------------------------------------------------------
void sys_idn_config_zn(double zn_k,double zn_ti,double zn_td)
{
	sys.zn_k=zn_k;
	sys.zn_ti=zn_ti;
	sys.zn_td=zn_td;
}

//---------------------------------------------------------------------------
// System identification - get Ziegler-Nichols tune parameters
//
// Requieres closed loop critical parameters to be set
// (=successfull system identification).
//---------------------------------------------------------------------------
void sys_idn_get_zn(double *k,double *ti,double *td)
{
	if(k)
		*k=sys.zn_k*sys.kp[0];
	if(ti)
		*ti=sys.zn_ti*sys.kp[1];
	if(td)
		*td=sys.zn_td*sys.kp[1];
}





